#! /usr/bin/env perl

package Smash::Analyses::Comparative::RefGenome;
use strict;
use warnings;
use Smash::Core;
use Smash::Global qw(:all);
use Smash::Utils::Taxonomy qw(:all);
use Smash::Utils::MatrixIO qw(read_R_matrix write_R_matrix zero_fill_matrix get_column_labels);

our @ISA = qw(Smash::Analyses::Comparative);

my $READ_PHYLO_MAP;
my $AVERAGE_GENOME_SIZE;
my %Species2Tax = ();
my %Sequence2Tax = ();
my %GenomeSize   = ();
my %SmashTax2NCBITax  = ();
my %ReadCount    = ();

sub init {
	my $this = shift;
	$this->SUPER::init();
	$this->{TYPE} = "refgenome";

	Smash::Utils::Taxonomy::init();

	# if localdb is used, then dont connect to database

# There is a problem in estimating genome size. Some times some genomes
# are sequenced multiple times, and they even use the same definition field.
# But this procedure used to just add all the sizes together for each genome,
# which is plain wrong. However, since some have multiple chromosomes, you
# should not use average of all either! The compromise is to only use one
# value per definition, so if the same sequence was sequenced twice then only
# one is counted. To make it reproducible, the first query is sorted by 
# sequence_id.
# CAVEAT: If there are two WGS genomes with the same tax_id and using different
#         assemblies, and hence different contig/scaffold names, this cannot be
#         solved. This must be handled manually!

	if (!$this->is_local) {
		my $smash = new Smash::Core();
		$smash->init();
		my $dbh   = $smash->get_refgenomedb_handle();

		# init seq2tax
		# init genome size

		my %Size;
		my $seq_sth = $dbh->prepare_cached("SELECT taxonomy_id, sequence_id, definition, length FROM sequence ORDER BY sequence_id");
		$seq_sth->execute();
		while (my ($tax_id, $seq_id, $definition, $length) = $seq_sth->fetchrow_array()) {
			$Sequence2Tax{$seq_id} = $tax_id;
			$Size{$tax_id}{$definition} = $length;
		}
		foreach my $tax_id (keys %Size) {
			$GenomeSize{$tax_id} = 0;
			foreach my $definition (keys %{$Size{$tax_id}}) {
				$GenomeSize{$tax_id} += $Size{$tax_id}{$definition};
			}
		}

		$smash->close_refgenomedb_handle();
		$smash->finish();
	} else {
		my $dbfile = $this->{LOCAL_DB_FILE};
		open(DB, "<$dbfile") || die "Cannot open $dbfile: $!";
		while (<DB>) {
			chomp();
			next if m/^#/;
			my @w = split(/\t/);
			my ($tax_id, $seq_id, $seq_length, $ncbi_tax_id) = ($w[0], $w[1], $w[4], $w[9]);
			$Sequence2Tax{$seq_id} = $tax_id;
			$GenomeSize{$tax_id} += $seq_length;
			$SmashTax2NCBITax{$tax_id} = $ncbi_tax_id;
		}
	}

	# estimate average genome size
	# used to be 3539003

	my $total = 0;
	my @sizes = values %GenomeSize;
	$total += $_ foreach @sizes;
	$AVERAGE_GENOME_SIZE = int($total/@sizes+0.5);
	print STDERR "Avg. genome size = $AVERAGE_GENOME_SIZE\n";
}

sub attach_sample {
	my $this       = shift;
	my $metagenome = shift || die "attach_sample needs metagenome";
	my $label      = shift || die "attach_sample needs label";
	my $smash      = $this->smash;

	my %options;

	# if localdb is used, then dont connect to database

	if (!$this->is_local) {
		%options = (METAGENOME => $metagenome);
	}
	$smash->finish() if $smash;
	$smash = new Smash::Core(%options);
	$smash->init();
	$this->{SMASH} = $smash;

	# for local dbfile, we skipped setting METAGENOME
	# if we didnt, then SMASH will try to connect to the database.
	# now we fix this by explicitly setting METAGENOME so that the next steps work

	$smash->{METAGENOME} = $metagenome unless $smash->metagenome; 

	$this->{LABEL} = $label;
	push(@{$this->{LABELS}}, $label);
}

sub get_data_count {
	my $this  = shift;
	my $metagenome = shift;

	# if local dbfile, just look it up in the hash

	return $ReadCount{$metagenome} if ($this->is_local);

	my $dbh   = $this->smash->get_db_handle();
	my $sth   = $dbh->prepare_cached("SELECT COUNT(DISTINCT template_id) FROM readinfo r INNER JOIN library l USING (library_id) INNER JOIN sample s USING (sample_id) WHERE metagenome_id=? AND r.length >= ?");
	$sth->execute($metagenome, 100);
	my ($count) = $sth->fetchrow_array();
	$sth->fetchrow_array();
	return $count;
}

# convert to abundance by normalizing by genome size

sub convert_to_abundance {

	my $this = shift;
	my $feature_count = shift;

	my $samples = get_column_labels($feature_count);
	my $data_count = $this->data_count;

	# Get unmapped counts

	foreach my $label (@$samples) {
		my $sum = 0;
		foreach my $tx (keys %$feature_count) {
			$sum += ($feature_count->{$tx}{$label} || 0);
		}
		if ($this->normalize ne "hits" && !defined($feature_count->{-1}->{$label})) {
			$feature_count->{-1}->{$label} = ($data_count->{$label}-$sum);
		}
	}

	zero_fill_matrix($feature_count);

	# record raw count in case of replicate

	if (defined($this->replicate)) {
		my $file_name = sprintf("%s.feature_vector.rawcount.%s.%s.%s", $this->name, $this->type, $this->refdb, $this->replicate);
		write_R_matrix($file_name, $feature_count);
	}

	# convert mapped counts to coverage, which is a proxy for number of individuals

	foreach my $tx (grep {$_ >= 0} keys %$feature_count) {
		my $genome_size = $GenomeSize{$tx} || die "Genome size unknown for $tx!";
		foreach my $label (@$samples) {
			my $c = $feature_count->{$tx}{$label} || 0;
			$feature_count->{$tx}{$label} = $c/$genome_size;
		}
	}

	# unmapped counts to coverage using avg genome length in refgenome set

	map {$feature_count->{-1}{$_} /= $AVERAGE_GENOME_SIZE if $feature_count->{-1}{$_};} @$samples;


	# BEG PCA hack

	#foreach my $tx (-1, 1350, 186813, 1678, 189330, 209, 1301, 46254, 1716, 1357, 1253) {
	#	delete $feature_count->{$tx};
	#}

	# END PCA hack
	return;
}

sub get_ncbi_tax_id {
	my $this = shift;
	my $tax_id = shift;

	# if local dbfile, just look it up in the hash

	return ($SmashTax2NCBITax{$tax_id} || $tax_id) if ($this->is_local);

	# otherwise, do the tedious DB lookup

	my $dbh = $this->smash->get_refgenomedb_handle();

	my $sth = $dbh->prepare_cached("SELECT source, display_tax_id FROM taxonomy WHERE taxonomy_id=?");
	$sth->execute($tax_id);
	my ($source, $ncbi_tax_id) = $sth->fetchrow_array();
	$sth->finish();

	$ncbi_tax_id = $tax_id unless defined $ncbi_tax_id; #  if not in table, has been processed elsewhere
	return $ncbi_tax_id;
}

sub merge_feature_vectors_just_in_case {
	my $this = shift;
	my $feature_count = shift;
	my $merge_level   = shift;
	my $rank_as_text  = shift;

	my $MergedFeatureCount;
	my $labels = get_column_labels($feature_count);

	$this->init_smash() unless defined $this->smash;

	# open it here, so that get_ncbi_tax_id doesnt have to open it

	$this->smash->get_refgenomedb_handle();

	my @ids = grep {$_ >= 0} keys %$feature_count; # skip unmapped = -1 here

	if ($this->tree =~ /NCBI/i) {
		foreach my $tax_id (@ids) {
			# the following line is obsolete since refgenome freeze4
			# my $ncbi_tax_id = $this->get_ncbi_tax_id($tax_id);
			my $ncbi_tax_id = $tax_id;
			my $new_id      = get_ncbi_taxonomic_rank($ncbi_tax_id, $merge_level);
			my $rank = $new_id;
			if ($rank_as_text) {
				$rank = $NCBITax2NCBIName{$rank};
				$rank =~ s/\s/_/g;
			}
	#print STDERR "$ncbi_tax_id\t$rank\n";
			foreach my $label (@$labels) {
				$MergedFeatureCount->{$rank}->{$label} += ($feature_count->{$tax_id}{$label}||0);
			}
		}
	} elsif ($this->tree =~ /bergey/i) {

my %SMASH2RDP;
open(SMASH2RDP, "<remap.txt");
while (<SMASH2RDP>) {
	chomp();
	next if m/^\s*#/;
	my ($smash_id, $ncbi_id, $rdp_id) = split(/\t/);
	$SMASH2RDP{$smash_id} = $rdp_id;
}
close(SMASH2RDP);

		TAX_ID:foreach my $tax_id (@ids) {
			my $rdp_tax_id = $SMASH2RDP{$tax_id};
if (!defined($rdp_tax_id)) {
	warn "$tax_id has no match\n";
	next TAX_ID;
}
			my $rank = get_bergey_taxonomic_rank($rdp_tax_id, $merge_level);
			if ($rank_as_text) {
				$rank = $BergeyTax2BergeyName{$rank};
				$rank =~ s/\s/_/g;
			}
#print STDERR "$rdp_tax_id\t$rank\n";
			foreach my $label (@$labels) {
				$MergedFeatureCount->{$rank}->{$label} += ($feature_count->{$tax_id}{$label}||0);
			}
		}
	}

	# handle unmapped
	foreach my $label (@$labels) {
		$MergedFeatureCount->{-1}->{$label} += ($feature_count->{-1}{$label}||0);
	}

	$this->smash->close_refgenomedb_handle();
	return $MergedFeatureCount;

=begin REMOVED_ELSE_PART

		my %Color = ("NCBI" => "#808284", "BROAD" => "#0000FF", "WUGSC" => "#006E3A", "BCM" => "#00ADEF", "JCVI" => "#8E191B", "SC" => "#F26522", "MIX" => "#8CC63F");
		my %Source; # tax2src mapping
		my %Display;# tax2display mapping
		my %DisplayIds; # display list

		foreach my $tax_id (grep {$_ >= 0} keys %$feature_count) {
			$sth->execute($tax_id);
			my ($source, $display_id) = $sth->fetchrow_array();
			# Keep track of tax_id and display_id - we need them later
			$Display{$tax_id} = $display_id;
			$DisplayIds{$display_id} = 1;
			$Source{$tax_id} = $source;
		}

		open(CENTERS, ">$opt_tag.$opt_analysis.$refdb.centers.dataset") || die "Cannot open $opt_tag.$opt_analysis.$refdb.centers.dataset: $!";
		foreach my $tax_id (keys %DisplayIds) {
			# Get all tax ids that map to this display_id
			my @genomes = grep {$Display{$_} eq $tax_id} keys %Display;
			foreach my $label (values %Labels) {
				$MergedFeatureCount->{$tax_id}->{$label} = sum_array(map {$feature_count->{$_}{$label}} @genomes);
			}

			# See how many centers are there
			my %buf = map {$Source{$_} => 1} @genomes;
			my @centers = keys %buf;
			if (scalar(@centers) == 1) {
				print CENTERS join("\t", $tax_id, $Color{$centers[0]}, $centers[0])."\n";
			} else {
				print CENTERS join("\t", $tax_id, $Color{MIX}, "MIX")."\n";
			}
		}
		close(CENTERS);

=cut

}

sub analyze_one_sample {
	use File::Temp;
	my $this              = shift;
	my ($bootstrap_label) = @_;
	my $label             = $this->label;
	my $identity          = $this->identity;
	my $feature_count     = $this->feature_count;
	my $smash             = $this->smash;
	my $metagenome        = $smash->metagenome;
	my $blast_file        = $metagenome.".".$this->refdb.".blastn";

	my %Template;

	# if this is not local, check for paired end data from the database

	if (!$this->is_local) {
		my ($id, $template);
		my $dbh = $smash->get_db_handle;
		my $sth = $dbh->prepare("SELECT read_id, template_id FROM readinfo WHERE read_id LIKE ?");
		$sth->execute("$metagenome.%");
		$sth->bind_columns(\$id, \$template);
		while ($sth->fetch()) {
			$Template{$id} = $template;
		}
	}

	my $db = new File::Temp();
	my $dbname = $db->filename;
	
	my $dbh = DBI->connect("DBI:SQLite:dbname=$dbname.sqlite", "", "", {AutoCommit => 0, RaiseError => 1}) || die "Couldn't connect to database: " .DBI->errstr;

	$dbh->do("CREATE TABLE templates(template_id VARCHAR(255) NOT NULL, taxonomy_id INTEGER NOT NULL)");
	$dbh->do("CREATE TABLE read_ref(read_id VARCHAR(255) NOT NULL, taxonomy_id INTEGER NOT NULL)");

	# Read the refgenome blasts

	{
		my $sth = $dbh->prepare("INSERT INTO templates(template_id, taxonomy_id) VALUES(?, ?)");
		my $sth2 = $dbh->prepare("INSERT INTO read_ref(read_id, taxonomy_id) VALUES(?, ?)");
		open(BLAST, "<$blast_file") || die "Cannot open $blast_file: $!";
		HSP:while (<BLAST>) {
			chomp;
			my ($read, $ref_sequence, $pi) = (split(/\s+/))[0,1,10];
			next HSP unless $pi > $identity;
			if ($ref_sequence =~ /ref\|([AN]C_[0-9]+)\./) {
				$ref_sequence = $1;
			}
			my ($tax_id) = $Sequence2Tax{$ref_sequence};
			defined($tax_id) || die "No match for $ref_sequence\n";

			$sth->execute($Template{$read} || $read, $tax_id); # for non paired-end, template is the same as read
			$sth2->execute($read, $tax_id);
		}
		close(BLAST);
	}

	# create index to make it faster

	$dbh->do("CREATE INDEX idx_t1 ON templates(template_id, taxonomy_id)");

	# we have a redundant list of (template, tax) now.

	# this is skippped now.
	# uncommenting this will treat single end and paired end mappings to a taxon the same. (fwd+rev, fwd only, rev only all get the same weight).
	#$dbh->do("CREATE TABLE templates_filtered(template_id VARCHAR(255), taxonomy_id INTEGER, CONSTRAINT pk_tf1 PRIMARY KEY(template_id, taxonomy_id))");
	#$dbh->do("INSERT INTO templates_filtered(template_id, taxonomy_id) SELECT DISTINCT template_id, taxonomy_id FROM templates");

	# Let's get the number of mappings per template, since we need that for sharing across taxa later

	$dbh->do("CREATE TABLE template_mapping_count(template_id VARCHAR(255) PRIMARY KEY, map_count INTEGER NOT NULL)");
	$dbh->do("INSERT INTO template_mapping_count(template_id, map_count) SELECT template_id, COUNT(*) AS map_count FROM templates GROUP BY template_id");

	# Let us now get a (template, tax, fraction) from them all
	# each such k-tuple give the fraction of template assigned to that tax. sum(fraction) over tax for each template must be 1.

	$dbh->do("CREATE TABLE template_tax_map(template_id VARCHAR(255) NOT NULL, taxonomy_id INT NOT NULL, count FLOAT(6,4), CONSTRAINT pk_tm1 PRIMARY KEY(template_id, taxonomy_id))");
	$dbh->do("INSERT INTO template_tax_map(template_id, taxonomy_id, count) SELECT t.template_id, taxonomy_id, SUM(1.0/map_count) FROM templates t INNER JOIN template_mapping_count tmc ON t.template_id=tmc.template_id GROUP BY t.template_id, taxonomy_id");

	$dbh->do("CREATE INDEX idx_r1 ON read_ref(read_id, taxonomy_id)");
	$dbh->do("CREATE INDEX idx_r2 ON read_ref(taxonomy_id)");
	$dbh->do("CREATE TABLE read_ref_map(read_id VARCHAR(255), taxonomy_id INTEGER, confidence FLOAT, CONSTRAINT pk_tf1 PRIMARY KEY(read_id, taxonomy_id))");
	{
		$this->smash->get_refgenomedb_handle();
		my %TaxMaps;
		my $sth = $dbh->prepare("SELECT DISTINCT taxonomy_id FROM read_ref");
		$sth->execute();
		while (my ($tax_id) = $sth->fetchrow_array()) {
			# this line is obsolete from refgenome freeze 4, since tax_ids are correct
			#my $new_id = $this->get_ncbi_tax_id($tax_id);

			my $new_id = $tax_id;
			#$TaxMaps{$tax_id} = get_taxonomic_rank($new_id, $this->level);
			$TaxMaps{$tax_id} = $new_id;
		}

		my %ReadRef;
		my $sth2 = $dbh->prepare("SELECT read_id, taxonomy_id FROM read_ref");
		$sth2->execute();
		while (my ($read, $tax_id) = $sth2->fetchrow_array()) {
			$ReadRef{$read}{$TaxMaps{$tax_id}} = 1;
		}
		foreach my $read (keys %ReadRef) {
			my @taxa = keys %{$ReadRef{$read}};
			foreach my $tax_id (@taxa) {
				printf "%s\t%d\t%f\n", $read, $tax_id, 1/@taxa;
			}
		}
		$this->smash->close_refgenomedb_handle();
	}

	$dbh->commit();

	my $assigned = 0;
	{
		my ($tax, $count);
		my $sth = $dbh->prepare("SELECT taxonomy_id, SUM(count) FROM template_tax_map GROUP BY taxonomy_id");
		$sth->execute();
		$sth->bind_columns(\$tax, \$count);
		while ($sth->fetch()) {
			$feature_count->{$tax}->{$label} = $count;
			$assigned += $count;
		}
	}
	#$sth->finish();
	#$dbh->do("DROP TABLE taxonomy_map");

	$dbh->do("DROP TABLE template_tax_map");
	$dbh->do("DROP TABLE template_mapping_count");
	$dbh->do("DROP TABLE templates");
	$dbh->commit();
	$dbh->disconnect();
	undef $dbh;
	unlink "$dbname.sqlite";
	#print STDERR "multihits in $label: $multihits\n";

	return $assigned;
}

1;
