############################################
#  GeneMark gene prediction program handler
############################################

package Smash::Analyses::GenePredictor::GeneMark;
use base "Smash::Analyses::GenePredictor";
use strict;

=head1 NAME

Smash::Analyses::GenePredictor::GeneMark - Implementation of GeneMark gene prediction software pipeline

=head1 SYNOPSIS

	my $instance = Smash::Analyses::GenePredictor::GeneMark->new
				(
					'NAME'        => $label,
					'OUTPUT_DIR'  => $output_dir,
					'PREDICTOR'   => 'GeneMark',
					'FASTA_FILE'  => $fasta,
					'SELF_TRAIN'  => 0,
					'VERSION'     => "2.6p"
				);
	$instance->run();

=head1 DESCRIPTION

GeneMark requires a license key file that must be obtained from the distributors of the software. This file must be available as 
F<B<pkg_dir>/gm_key> (See L<Smash::Analyses/"init_software_details"> for explanation of what B<pkg_dir> is).

=cut

############################################
# Constructor
############################################

sub new1 {
	my $class  = shift;
	my %params = @_;
	my $self   = $class->SUPER::new(@_);
	$self->{PREDICTOR} = "GeneMark";
	return $self;
}

=head1 FUNCTIONS

=head2 Mandatory and conditionally mandatory functions

=over 4

=item B<is_trainable>

returns C<1>.

=item B<train_min>

returns C<200000>.

=item B<get_self_trained_settings>

Trains a prokaryotic gene model using the input fasta 
file and returns the command line option to use that 
model.

=item B<get_generic_settings>

Chooses a heuritic model with the right translation 
table and the GC content of the input sequence and 
returns the command line option to use that  model.

=item B<get_command_line>

Makes a command line with the model passed.

=item B<parameter_settings>

describes the rule based creation of parameter set
for commandline.

=item B<create_gene_gff>

=back

=cut

sub is_multifasta_compatible {return 0;}
sub is_trainable {return 1;}
sub train_min    {return 200000;}

sub train_combo_parameter {
	my $organism = shift;
	my $option;
	if ($organism eq "bacteria") {
		$option   = "--prok";
	} elsif ($organism eq "virus") {
		$option = "--virus";
	} elsif ($organism eq "eukaryote") {
		$option = "--euk";
	} elsif ($organism eq "mold_mitochondrial") {
		$option = "--gcode 4 --shape circular";
	}
	return $option;
}

sub get_self_trained_settings {
	my $this   = shift;
	my $name   = $this->name;
	my $output_dir = $this->output_dir;
	my $option = train_combo_parameter($this->organism);
	my $command = sprintf("cd $output_dir && %s/gmsn.pl $option --name=$name --maxitr=100 %s", $this->pkg_dir, $this->fasta_filename);
print "$command\n";
	$this->execute($command);
	return "-r -m $output_dir/${name}_hmm_combined.mod";
}

############################################
# Construct the model filename from 
# GC content and translation table
############################################

sub get_generic_settings {
	my $this        = shift;
	my $input_file  = shift;
	my $trans_table = $this->translation_table;
	my $modeldir    = $this->paramdir;

	my $model = "-m $modeldir/heu_${trans_table}_<__GC__>.mod";

	return $model;
}

sub get_command_line {
	my $this        = shift;
	my $input_file  = shift;
	my $output_file = shift;
	my $model_arg   = shift;
	my $pkg_dir     = $this->pkg_dir;
	my $exe         = $this->exe;

	if ($model_arg =~ /<__GC__>/) {
		open(FASTA, "<$input_file") || die "Cannot open $input_file: $!";
		my $fasta    = new FAlite(\*FASTA);
		my $gcp;
		if (my $entry = $fasta->nextEntry) {
			$gcp  = $this->get_gc_percent($entry->seq, 30, 70);
		}
		close(FASTA);
		$gcp   = int($gcp + 0.5); # quick hack for round
		$model_arg =~ s/<__GC__>/$gcp/;
	}


	my $command       = "cd $pkg_dir && ./$exe $model_arg -o $output_file $input_file";

	# Used this for E. coli complete genomes
	#my $command       = "cd $pkg_dir && ./$exe -m $modeldir/ecoli_r.mod -r -o $output_file $input_file";

	return $command;
}

sub parameters {
	my $this        = shift;
	my $trans_table = $this->translation_table;
	my $organism    = $this->organism;
	if ($this->self_train == 1) {
		return "-m self_train_$organism -r";
	} else {
		return "-m heu_${trans_table}_gc";
	}
}

sub parameter_settings {
	my $this        = shift;
	my $trans_table = $this->translation_table;
	my $organism    = $this->organism;
	my $train_min   = $this->train_min;
	my $settings    = "";
	if ($this->self_train == 1) {
		$settings .= "'-m self_train_$organism -r' for length > $train_min; ";
		$settings .= "'-m heu_${trans_table}_gc' otherwise";
	} else {
		$settings .= "'-m heu_${trans_table}_gc'";
	}
	return $settings;
}

############################################
# Process GeneMark's raw output
############################################

sub create_gene_gff {
	my $this          = shift;
	my $GFF_FH        = shift;
	my $def           = shift;
	my $pred_file     = shift;
	my $genepred      = $this->genepred;
	my $assembly      = $this->assembly;
	my $IN_FH;

	open($IN_FH, "<$pred_file") || die "Cannot open $pred_file: $!";

	# Skip the header section
	PRESCAN:while (<$IN_FH>) {
		if ($_ =~ /^\s*#\s*Length\s*$/) {
			last PRESCAN;
		}
	}

	# Now to the predictions
	####################
	# FIX THE FRAME
	# GENEMARK always lists complete codons, so frame is always 0
	###################

	$, = "\t";
	LINE:while (<$IN_FH>) {
		my $line = $_;

		# Skip empty predictions
		if ($line =~ /^-+$/) {
			last LINE;
		}

		# Parse the gene lines
		if ($line =~ /^\s*(\d+)\s+(\+|-)\s+(<*)(\d+)\s+(>*)(\d+)\s+/) {
			my ($gene_id, $strand, $langle, $start, $rangle, $end) = ($1, $2, $3, $4, $5, $6);
			my $incomplete5  = 0;
			my $incomplete3  = 0;
			my ($start_codon, $stop_codon);
			if ($langle eq "<") {
				$incomplete5 = 1;
			}
			if ($rangle eq ">") {
				$incomplete3 = 1;
			}
			if ($strand eq "+") {
				$start_codon = ($incomplete5 == 1)?"no":"yes";
				$stop_codon  = ($incomplete3 == 1)?"no":"yes";
			} else {
				$stop_codon  = ($incomplete5 == 1)?"no":"yes";
				$start_codon = ($incomplete3 == 1)?"no":"yes";
			}
			if ($def =~ /${assembly}\.(.*)/) {
				$gene_id = "$genepred.$1.G$gene_id";
			} else {
				$gene_id = "$genepred.$def.G$gene_id";
				#die "Invalid contig id: $def";
			}
			print $GFF_FH $def, "$this", "CDS", $start, $end, ".", $strand, ".", "gene_id \"$gene_id\"; start_codon \"$start_codon\"; stop_codon \"$stop_codon\";\n";
		} else {
			die "Incorrect format found in GeneMark output: $line";
		}
	}
	$, = "";
}

=head2 Suggested functions

=over 4

=item B<min_length>

returns C<41>.

=back

=cut

sub min_length   {return 41;}

=head2 Local functions

These functions are not known outside of the scope of GeneMark.

=over 4

=item B<check_license_key>

Checks for the GeneMark software license key and copies it to the user's home directory. This is required for GeneMark to
run properly. The license key file should be available as F<B<pkg_dir>/gm_key>. See L</DESCRIPTION> for more details.

=item B<paramdir>

Directory where the parameter files reside under B<pkg_dir> for GeneMark.

=back

=cut

sub exe          {return "gmhmmp";}

############################################
# Initialize pkg variables
############################################

sub init {
	my $this = shift;
	$this->SUPER::init();
	PLACES: foreach my $dir qw(modeldir heuristic_mod) {
		my $candidate = $this->pkg_dir."/$dir";
		if (-d $candidate) {
			$this->set_paramdir($candidate);
			last PLACES;
		}
	}
	$this->check_license_key();
}

############################################
# check for the license key in the user's home directory, or
# changed on 12.04.2009: version/key mismatches are possible.
#                        so dont check, just copy!
# copy the license key from smash directory to the user's home directory
############################################

sub check_license_key {
	use File::Copy;
	use Env qw(HOME);
	my $this = shift;
	my $pkg_dir = $this->pkg_dir;
	my $exe     = $this->exe;
	my $output  = `$pkg_dir/$exe 2>&1`;
	if ($output =~ /License key .* not found/si ||
	    $output =~ /Segmentation fault/si ||
	    $output =~ /Your .* license period has ended/si) {
			copy("$pkg_dir/gm_key", "$HOME/.gm_key") || die "Cannot copy license key from $pkg_dir/gm_key: $!";
	}
}

############################################
# GeneMark parameter file locations
############################################

sub paramdir {shift->{PARAMDIR}}
sub set_paramdir {
	my $this = shift;
	$this->{PARAMDIR} = shift;
}

############################################
# Return the command line to run GeneMark 
# on a sequence
# If $model is not defined, use heuristics
############################################

1;

__END__

=head1 DEPENDENCIES

L<File::Copy>

=head1 AUTHOR

Mani Arumugam (arumugam@embl.de, http://www.embl.de)

=head1 COPYRIGHT

Copyright (C) 2008-2009. Mani Arumugam and EMBL. All rights reserved.

=cut
