package Smash::Databases::MetaGenomeDB::AssemblyLoader;

use strict;
use warnings;

our @ISA = qw(Smash::Databases::MetaGenomeDB::Loader);

use File::Path;
use Smash::Utils::GFFlite;

############################################
# Constructor
############################################

my $dbh;
my $c2r_sth;
my $s2c_sth;
my $contig_sth;
my $scaffold_sth;
my $PROGRESS;

sub init {
	my $this = shift;
	$this->parse_config();
	my ($collection, $metagenome) = $this->parse_assembly_id($this->assembly);
	$this->{COLLECTION} = $collection;
	$this->{METAGENOME} = $metagenome;
	$this->SUPER::init();

	# Database stuff

	$dbh          = $this->get_db_handle;
	$c2r_sth      = $dbh->prepare('INSERT INTO contig2read(contig_id, read_id, start, end, strand) VALUES(?, ?, ?, ?, ?);') || die "Cannot prepare statement: ".$dbh->errstr;
	$contig_sth   = $dbh->prepare('INSERT INTO contig(external_id, read_count, length, assembly_id) VALUES(?, ?, ?, ?);') || die "Cannot prepare statement: ".$dbh->errstr;
	$s2c_sth      = $dbh->prepare('INSERT INTO scaffold2contig(scaffold_id, contig_id, start, end, strand) VALUES(?, ?, ?, ?, ?);') || die "Cannot prepare statement: ".$dbh->errstr;
	$scaffold_sth = $dbh->prepare('INSERT INTO scaffold(external_id, contig_count, length, assembly_id) VALUES(?, ?, ?, ?);') || die "Cannot prepare statement: ".$dbh->errstr;
	$PROGRESS = \*STDERR;
	select($PROGRESS); $| = 1; select(STDOUT);
}

sub finish {
	my $this = shift;
	$c2r_sth->finish();
	$s2c_sth->finish();
	$contig_sth->finish();
	$scaffold_sth->finish();
	$dbh->commit();
	$this->SUPER::finish();
}

sub unload_db {
	my $this = shift;

	# Locations etc

	my $assembly    = $this->assembly;
	my $assembly_id = $this->get_id_by_name("assembly", $assembly);
	my $dbh;

	# Remove from MC* database
	print "Removing entries from scaffold, scaffold2contig, contig and contig2read\n";

	$dbh = $this->get_db_handle;

	{
		####
		# the following two deletes are skipped, since I added an "ON DELETE CASCADE" constraing on scaffold2contig and contig2read
		####

		#$sth = $dbh->prepare('DELETE s2c.* FROM scaffold2contig s2c LEFT JOIN scaffold s USING (scaffold_id) WHERE assembly_id=?') || die "Cannot prepare statement:".$dbh->errstr;
		#$sth->execute($assembly_id);
		#$sth = $dbh->prepare('DELETE c2r.* FROM contig2read c2r LEFT JOIN contig c USING (contig_id) WHERE assembly_id=?') || die "Cannot prepare statement:".$dbh->errstr;
		#$sth->execute($assembly_id);

		my $sth1 = $dbh->prepare('DELETE FROM scaffold WHERE assembly_id=?') || die "Cannot prepare statement:".$dbh->errstr;
		$sth1->execute($assembly_id);
		#$sth1->fetchrow_array(); # perl DBI shush
		$sth1 = undef;
		my $sth2 = $dbh->prepare('DELETE FROM contig WHERE assembly_id=?') || die "Cannot prepare statement:".$dbh->errstr;
		$sth2->execute($assembly_id);
		#$sth2->fetchrow_array(); # perl DBI shush
		$sth2 = undef;
	}
	$dbh->commit();
}

sub wipe_out {
	my $this         = shift;
	my $assembly     = $this->assembly;
	my $assembly_dir = $this->assembly_dir($assembly);
	my $assembly_id  = $this->get_id_by_name("assembly", $assembly);
	my $dbh;

	print "Removing entry from SmashDB.assembly\n";
	$dbh = $this->get_smashdb_handle;
	{
		my $sth = $dbh->prepare('DELETE FROM assembly WHERE assembly_id=?');
		$sth->execute($assembly_id);
		#$sth->fetchrow_array(); # perl DBI shush
		$sth = undef;
	}
	$dbh->commit();

	print "Removing files from $assembly_dir\n";
	rmtree($assembly_dir);
}

############################################
# Load reads and xml information into db
############################################

sub run {
	use Fcntl qw(:seek);
	my $this = shift;
	
	# Locations etc

	my $assembly     = $this->assembly;
	my $assembly_dir = $this->assembly_dir($assembly);
	my $contig2read  = "$assembly_dir/$assembly.contig2read.gff";
	my $scaf2contig  = "$assembly_dir/$assembly.scaf2contig.gff";

	my $assembly_id  = $this->get_id_by_name("assembly", $assembly);
	if (!$assembly_id) {
		die "Assembly $assembly does not exist in Smash. Please check your assembly run!";
	}

	my %SeqLength; # Track contig/scaf length
	my %SeqCount; # Track contig/scaf featurecount
	my %ContigExternal2Internal; # for contig external to internal mapping
	my %ScaffoldExternal2Internal; # for scaffold external to internal mapping

	# Parse contig2read GFF

	print $PROGRESS "Contig2Read:\n";

	%SeqLength = ();
	%SeqCount = ();

	print $PROGRESS "\tReading GFF file ...";

	open(CONTIG2READ, "<$contig2read") || die "Cannot open $contig2read: $!";
	my $gff   = new Smash::Utils::GFFlite(\*CONTIG2READ);
	while (my $f = $gff->nextFeature) {
		my $name = $f->seqname; 
		$SeqLength{$name} = $f->score; 
		$SeqCount{$name}++;
	}
	print $PROGRESS " done\n";

	# Process contig2read maps

	print $PROGRESS "\tLoading into database ...";
	foreach my $external_id (keys %SeqLength) {
		$contig_sth->execute($external_id, $SeqCount{$external_id}, $SeqLength{$external_id}, $assembly_id); # We use score field for length
		$ContigExternal2Internal{$external_id} = $this->last_db_insert_id;
	}
	print $PROGRESS ".";

	# read the contig2read gff file again. Don't reopen, since it is a waste. Just rewind
	# but rebless a new GFFlite object!

	seek(CONTIG2READ, 0, SEEK_SET);
	$gff   = new Smash::Utils::GFFlite(\*CONTIG2READ);
	while (my $feature = $gff->nextFeature) {
		my $read_id = $feature->get_attribute("read");
		$c2r_sth->execute($ContigExternal2Internal{$feature->seqname}, $read_id, $feature->location->[0]->start, $feature->location->[0]->end, $feature->strand);
	}
	close(CONTIG2READ);
	print $PROGRESS " done\n";

	# Parse scaf2contig GFF

	print $PROGRESS "Scaf2Contig:\n";

	%SeqLength = ();
	%SeqCount = ();

	print $PROGRESS "\tReading GFF file ...";

	open(SCAF2CONTIG, "<$scaf2contig") || die "Cannot open $scaf2contig: $!";
	$gff   = new Smash::Utils::GFFlite(\*SCAF2CONTIG);
	while (my $f = $gff->nextFeature) {
		my $name = $f->seqname; 
		$SeqLength{$name} = $f->score; 
		$SeqCount{$name}++;
	}
	print $PROGRESS " done\n";

	# Process scaf2contig maps

	print $PROGRESS "\tLoading into database ...";
	foreach my $external_id (keys %SeqLength) {
		$scaffold_sth->execute($external_id, $SeqCount{$external_id}, $SeqLength{$external_id}, $assembly_id); # We use score field for length
		$ScaffoldExternal2Internal{$external_id} = $this->last_db_insert_id;
	}
	print $PROGRESS ".";

	seek(SCAF2CONTIG, 0, SEEK_SET);
	$gff   = new Smash::Utils::GFFlite(\*SCAF2CONTIG);
	while (my $feature = $gff->nextFeature) {
		my $contig_id = $ContigExternal2Internal{$feature->get_attribute("contig")};
		$s2c_sth->execute($ScaffoldExternal2Internal{$feature->seqname}, $contig_id, $feature->start, $feature->end, $feature->strand);
	}
	print $PROGRESS " done\n";
}

1;
