package Smash::Utils::Taxonomy;

use strict;
use warnings;
use Smash::Utils::HTML qw(get_html_page post_html_page strip_html get_attribute retrieve_remote_file);
use Smash::Global qw(:all);

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(%Rank2Num $Num2Rank);

# NCBI related stuff

push(@EXPORT_OK, qw(get_local_taxonomy_for_id get_ordered_local_taxonomy_for_id get_full_lineage_bergey get_full_lineage_ncbi get_full_lineage_string));
push(@EXPORT_OK, qw(fix_lineage get_ncbi_taxonomic_rank get_taxonomy_for_complete_name get_taxonomy_for_token_set get_taxonomy_for_id));
push(@EXPORT_OK, qw(get_common_ancestor));
push(@EXPORT_OK, qw(%NCBITax2NCBIName %NCBITax2NCBIRank %NCBITax2NCBIParent %NCBIRankedName2NCBITax));

# RDP / Bergey related stuff

push(@EXPORT_OK, qw(get_bergey_taxonomic_rank bergey2ncbi_name_rank ncbi2bergey_name_rank));
push(@EXPORT_OK, qw(%BergeyTax2BergeyName %BergeyTax2BergeyRank %BergeyTax2BergeyParent %BergeyRankedName2BergeyTax $BergeyRoot));

our %EXPORT_TAGS = ('taxonomy' => [grep {$_ =~ /taxonomy/} @EXPORT_OK], 'all' => [@EXPORT_OK]);

# Ivica's tree parsing routine:

my %names;

# General

our %Rank2Num;
our %Num2Rank;

# NCBI

our %NCBITax2NCBIName;
our %NCBITax2NCBIRank;
our %NCBITax2NCBIParent;
our %NCBIRankedName2NCBITax;
our %Old2New; # hope it is one-to-one
our %New2Old; # hope it is one-to-one

# Bergey

our $BergeyRoot;
our %BergeyRankedName2BergeyTax;
our %BergeyTax2BergeyRank;
our %BergeyTax2BergeyName;
our %BergeyTax2BergeyParent;

# Big shots!

my $BACTERIA_ID;
my $ARCHAEA_ID;

=head1 NAME

Smash::Utils::Taxonomy - NCBI taxonomy related utility functions

=head1 SYNOPSIS

	use Smash::Utils::Taxonomy qw(:all);
	Smash::Utils::Taxonomy::init();

=head1 DESCRIPTION

Smash::Utils::Taxonomy provides several useful functions that are related to NCBI taxonomy ids.

=over 4

=item B<get_taxonomy_for_complete_name>

Performs a search using "complete name" mode on NCBI taxonomy website and
returns the taxonomy if found.

=item B<get_taxonomy_for_token_set>

Performs a search using "token set" mode on NCBI taxonomy website and
returns the taxonomy if found.

=item B<get_taxonomy_for_id>

Gets the taxonomy for given taxonomy id from NCBI taxonomy website.

=cut
	
sub get_taxonomy_for_complete_name {
	get_taxonomy_by_searching(shift, 1);
}

sub get_taxonomy_for_token_set {
	get_taxonomy_by_searching(shift, 3);
}

sub get_taxonomy_for_id {
	get_taxonomy_by_searching(shift, 5);
}

sub get_taxonomy_by_searching {
	my $srchtext = shift;
	my $srchmode = shift;
	my %taxonomy;
	my $name;
	my $tax_id;
	my $page;
	my $rank;
	my ($begin, $end);
	my $buf; 
	my %params = (name => $srchtext, srchmode => $srchmode, keep => 0, lvl => 3);

	###############################################
	# Post the request to NCBI Taxonomy Page
	###############################################

	$page = post_html_page("http://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi", \%params);

	###############################################
	# Get the required section
	###############################################

	if ($page !~ m#<em>Taxonomy ID:#) {
		($page) = ($page =~ m^<LI TYPE=circle>(<A.*?</A>)^);
		$page || return;
		my ($url) = get_attribute($page, "href");
		$url   = "http://www.ncbi.nlm.nih.gov$url";
		$page   = get_html_page($url);
	}
	($page) = ($page =~ m#<!--  the contents   -->(.*)</dl>#s);

	###############################################
	# Get NCBI taxonomy ID
	###############################################

	$buf   = $page;
	($tax_id) = ($buf =~ m^<em>Taxonomy ID:.*?</em>(.*?)<br>^s);
	($rank)   = ($buf =~ m^<em>Rank:.*?</em>(.*?)<br>^s);

	###############################################
	# Get NCBI standard name
	###############################################

	$buf   = $page;
	if ($buf =~ m#<h2>(.*)</h2>#s) {
		$name = strip_html($1);
	}

	###############################################
	# Get the full NCBI lineage
	###############################################

	my @pieces = ($buf =~ m#(<a ALT=.*?</a>)#sg);

	###############################################
	# Some pages have a slightly different format
	###############################################

	if (!@pieces) {
		@pieces = ($buf =~ m#(<A HREF="Taxonomy".*?</A>);#sg);
	}

	###############################################
	# Parse the lineage into a hash
	###############################################

	my $level = 1;
	%taxonomy = map {
		my (undef, $key) = m#ALT=(["'])([^\1]*?)\1#;
		my $val = strip_html($_);
		$key =~ s/ /_/g;
		if ($key eq "no_rank") {
			$key .= "_$level";
		}
		$level++;
		$key => $val
	} @pieces;
	$taxonomy{ncbi_name} = $name;
	$taxonomy{tax_id} = $tax_id;
	if ($rank ne "no rank" && !defined($taxonomy{$rank})) {
		$taxonomy{$rank} = $name;
	}
	return %taxonomy;
}

# map from bergey taxonomy to ncbi taxonomy

sub bergey2ncbi_name_rank {
	my ($name, $rank) = @_;

	# hacks

	if ($name eq "Escherichia/Shigella" && $rank eq "genus") {
		$name = "Escherichia";
	} elsif ($name eq "Parasutterella" && $rank eq "genus") {
		$rank = "subtribe";
	} elsif ($name eq "Marvinbryantia" && $rank eq "genus") {
		$name = "Bryantella";
	} elsif ($name eq "Gracilibacteraceae" && $rank eq "family") {
		$name = "Graciibacteraceae";
	} elsif ($name eq "Bacteria" && $rank eq "domain") {
		$rank = "superkingdom";
	} elsif ($name eq "Root" && $rank eq "norank") {
		$rank = "no rank";
		$name = "root";
	}

	my $ncbi_tax_id = $NCBIRankedName2NCBITax{$rank}{$name} || $NCBIRankedName2NCBITax{$rank}{"$name ($rank)"} || $NCBIRankedName2NCBITax{"no rank"}{$name};
	warn "WARNING: No NCBI match for $rank=$name\n" unless $ncbi_tax_id;

	return $ncbi_tax_id;
}

sub ncbi2bergey_name_rank {
	my ($name, $rank) = @_;

	# hacks

	if (($name eq "Escherichia" || $name eq "Shigella") && $rank eq "genus") {
		$name = "Escherichia/Shigella";
	} elsif ($name eq "Prochlorococcus" && $rank eq "genus") {
		$name = "GpIIa";
	} elsif ($name eq "Parasutterella" && $rank eq "subtribe") {
		$rank = "genus";
	} elsif ($name eq "Bryantella" && $rank eq "genus") {
		$name = "Marvinbryantia";
	} elsif ($name eq "Graciibacteraceae" && $rank eq "family") {
		$name = "Gracilibacteraceae";
	} elsif ($name eq "Bacteria" && $rank eq "superkingdom") {
		$rank = "domain";
	} elsif ($name eq "root" && $rank eq "no rank") {
		$rank = "norank";
		$name = "Root";
	}

	my $rdp_tax_id = $BergeyRankedName2BergeyTax{$rank}{$name};
	warn "WARNING: No RDP/Bergey match for $rank=$name\n" unless $rdp_tax_id;

	return $rdp_tax_id;
}

########################
# Get the Bergey taxonomy information
########################

sub parse_rdp_taxonomy_dump {
	my $this = shift;

	# parse tree

	my $smash     = new Smash::Core();
	   $smash->init();
	my ($rdp_dir) = $smash->software_dir("rdp_classifier", "current");
	   $smash->finish();
	my $tree_file = "$rdp_dir/training_data/classifier/bergeyTrainingTree.xml";
	open(TREE, "<$tree_file") || die "Cannot open $tree_file: $!";
	LINE: while (<TREE>) {
		chomp();
		if (!m/<TreeNode/) {
			next LINE;
		}
		my $name   = get_attribute($_, "name");
		my $rank   = get_attribute($_, "rank");
		my $tax    = get_attribute($_, "taxid");
		my $parent = get_attribute($_, "parentTaxid");

		# remove quotes

		$name =~ s/^['"]//;
		$name =~ s/['"]$//;
		$name =~ s/^&quot;//;
		$name =~ s/&quot;$//;

		# Store root

		if (lc($name) eq "root") {
			$BergeyRoot = $tax;
		}

		# If something's called just "Incertae Cedis XIII" or so, we prepend the previous rank to resolve ambiguity
		if ($name =~ /^Incertae/i) {
			$name = $BergeyTax2BergeyName{$parent}." $name";
		}
		$BergeyTax2BergeyName{$tax}   = $name;
		$BergeyTax2BergeyRank{$tax}   = $rank;
		$BergeyTax2BergeyParent{$tax} = $parent;

		$names{$tax} = $name;

		# check name clashes

		die "$rank:$name has two entries: $tax, $BergeyRankedName2BergeyTax{$rank}{$name}\n" if $BergeyRankedName2BergeyTax{$rank}{$name};
		$BergeyRankedName2BergeyTax{$rank}{$name}  = $tax;
	}
	close(TREE);

	# prefix ambiguous names

	my @ids = keys %BergeyTax2BergeyName;
	foreach my $id (@ids) {
		my $name = $BergeyTax2BergeyName{$id};
		if ($name =~ /^Gp[0-9IVX]+[a-z]?$/ || $name =~ /^Family [0-9IVX]+$/ || $name =~ /^Incertae Sedis [0-9IVX]+$/) {
			my $parent = $BergeyTax2BergeyName{$BergeyTax2BergeyParent{$id}};
			$name   = "$parent $name";
			$BergeyTax2BergeyName{$id} = $name;
		}
	}
}

sub get_pruned_tree {
	my ($system, $myroot, $list_ref) = @_;
	my $Tax2Parent;
	if (lc($system) eq "bergey") {
		$Tax2Parent = \%BergeyTax2BergeyParent;
		$myroot = $BergeyRoot unless defined $myroot;
	} else {
		$Tax2Parent = \%BergeyTax2BergeyParent;
		$myroot = $BergeyRoot unless defined $myroot;
	}
	my %genTree = prune_tree($list_ref, $Tax2Parent, $myroot);
	my $out_tree = dump_tree(\%genTree, 1,'', $myroot);
	$out_tree =~ s/,\)/)/g;
}

# Begin From Ivica

# dump_tree(tree_hash_ref, print_name?, tree_so_far, node_to_dump_from)

sub dump_tree {
  my $tree_ref= shift;
  my %genTree = %$tree_ref;
  my $use_name= shift;
  my $current = shift;
  my $id      = shift;
  my $int_prefix = '';
  $current .= "(";
  my @nkids = keys %{$genTree{$id}{'kids'}};
  foreach my $kID (@nkids) {
    my @gkids = keys %{$genTree{$kID}{'kids'}};
    if (not $gkids[0]) {
        if ($use_name) {
          $current .= $names{$kID} . ",";
        } else {
          $current .= $kID . ",";
        }
    } else {
      $current =  dump_tree($tree_ref, $use_name, $current, $kID);
      $current .= ',';
    }
  }
  if ($use_name) {
    $current .= ")" . $names{$id};
  } else {
    $current .= ")INT$id";
 }
 return $current;
}

sub prune_tree {
   my ($arg1, $arg2, $root) = @_;

   my @ids  = @{$arg1}; #selection
   my %tree = %{$arg2}; #simpler hash, ID -> parent
   my %genTree;
   
   foreach my $id (@ids) {
     if (defined $tree{$id}) {
       my $parID = $tree{$id};
       if ($parID == $root) {
         $genTree{"$parID"}{kids}{"$id"}++;
       } else {
         $genTree{"$parID"}{kids}{"$id"}++;
         #walk the tree up to the root
         while (1) {
           my $gparID = $tree{$parID};
           if ($gparID == $root) {
             $genTree{"$gparID"}{kids}{"$parID"}++;
             last;
           }
           $genTree{"$gparID"}{kids}{"$parID"}++;
           $parID = $gparID;
         }
       }
     } else {
       warn("Error: Cannot find ID $id in NCBI taxonomy files");
     }
   }
   return %genTree;
}

# End   From Ivica

########################
# Get the NCBI taxonomy information
########################

sub parse_ncbi_taxonomy_dumps {

	use Fcntl qw(:seek);

	# what to parse? only microbes, or every thing?

	my $what_to_parse = shift;
	my $only_bacteria_and_archaea = 0;
	if ($what_to_parse && $what_to_parse =~ /bacteria\/archaea/i) {
		$only_bacteria_and_archaea = 1;
	}

	# parse the ncbi dump files

	my $nodes_file = "$TAXONOMY_LOCAL_LOCATION/nodes.dmp";
	my $names_file = "$TAXONOMY_LOCAL_LOCATION/names.dmp";
	my $merge_file = "$TAXONOMY_LOCAL_LOCATION/merged.dmp";
	open(NODES, "<$nodes_file") || die "Cannot open $nodes_file: $!";
	while (<NODES>) {
		chomp();
		s/\t\|$//;
		my ($tax_id, $parent, $rank) = split(/\t\|\t/);
		$NCBITax2NCBIParent{$tax_id} = $parent;
		$NCBITax2NCBIRank{$tax_id}   = $rank;
	}
	close(NODES);

	# parse names.dmp

	open(NAMES, "<$names_file") || die "Cannot open $names_file: $!";

	# get $BACTERIA_ID and $ARCHAEA_ID

	PRESCAN: while (<NAMES>) {
		chomp();
		s/\t\|$//;
		my ($tax_id, $name, $uname, $class) = split(/\t\|\t/);
		if ($uname eq "Bacteria <prokaryote>" && $class eq "scientific name") {
			$BACTERIA_ID = $tax_id;
		}
		if ($name eq "Archaea" && $class eq "scientific name") {
			$ARCHAEA_ID = $tax_id;
		}
		last PRESCAN if ($BACTERIA_ID && $ARCHAEA_ID);
	}

	die "Could not find ids for Bacteria and Archaea!\n" if (!$BACTERIA_ID && !$ARCHAEA_ID);

	# reparse

	seek(NAMES, 0, SEEK_SET);
	NAME: while (<NAMES>) {
		chomp();
		s/\t\|$//;
		my ($tax_id, $name, $uname, $class) = split(/\t\|\t/);
		if ($only_bacteria_and_archaea && !is_bacteria_or_archaea($tax_id) && $name ne "unidentified") {
			next NAME;
		}
		my $rank = $NCBITax2NCBIRank{$tax_id};
		if ($class eq "synonym") {
			#warn "$rank: $name\n" if $NCBIRankedName2NCBITax{$rank}{$name};
			$NCBIRankedName2NCBITax{$rank}{$name} = $tax_id unless $NCBIRankedName2NCBITax{$rank}{$name};
			#printf "Rankedname2Tax{%s}{%s} = %d\n", $rank, $name, $tax_id;
		} elsif ($class eq "scientific name") {
			#warn "$rank: $name\n" if $NCBIRankedName2NCBITax{$rank}{$name};
			$NCBIRankedName2NCBITax{$rank}{$name} = $tax_id;
			if ($uname) {
				warn "$rank: $name\n" if $NCBIRankedName2NCBITax{$rank}{$uname};
				$NCBIRankedName2NCBITax{$rank}{$uname} = $tax_id;
			}
			$NCBITax2NCBIName{$tax_id} = $name;
		}
	}
	close(NAMES);

	open(MERGED, "<$merge_file") || die "Cannot open $merge_file: $!";
	while (<MERGED>) {
		chomp();
		s/\t\|$//;
		my ($old_id, $new_id) = split(/\t\|\t/);
		$Old2New{$old_id} = $new_id;
		$New2Old{$new_id} = $old_id;

		# Also give names to the old taxonomy_id's

		$NCBITax2NCBIName{$old_id} = $NCBITax2NCBIName{$new_id};
	}
	close(MERGED);
}

=item B<init>

=item B<get_full_lineage_ncbi>

Gets an array of the full taxonomy lineage as taxonomy ids. Starts from the lowest
level (given taxonomy id) and ends at kingdom.

=item B<get_local_taxonomy_for_id>

Same as C<get_taxonomy_for_id>, only gets the information from local NCBI taxonomy
dump files.

=item B<get_ordered_local_taxonomy_for_id>

Same as C<get_local_taxonomy_for_id>, but prepends the ordinal rank followed by
underscore so that the hash can be sorted using the ordinal rank.

=cut

sub update_files {
	use File::Temp qw(tempdir);
	use File::Copy;
	use File::Path;
	select(STDERR); $| = 1; select(STDOUT);
	my $filename = "taxdump.tar.gz";
	my $tmpdir = tempdir(CLEANUP => 1);
	print STDERR "NCBI Taxonomy repository:\n";
	print STDERR "Remote location: $NCBI_TAXONOMY_REMOTE_LOCATION\n";
	print STDERR "Local  location: $TAXONOMY_LOCAL_LOCATION\n";
	print STDERR "Retrieving remote files from remote location ...";
	system("wget -O $tmpdir/$filename $NCBI_TAXONOMY_REMOTE_LOCATION/$filename") == 0 || die "download failed: $!"; 
	print STDERR " done\n";
	print STDERR "Extracting files ...";
	system("cd $tmpdir && tar xfz $filename") == 0 || die "File extraction failed: $!";
	print STDERR " done\n";
	if (! -d "$TAXONOMY_LOCAL_LOCATION") {
		mkpath $TAXONOMY_LOCAL_LOCATION;
	}
	copy "$tmpdir/nodes.dmp", "$TAXONOMY_LOCAL_LOCATION/";
	copy "$tmpdir/names.dmp", "$TAXONOMY_LOCAL_LOCATION/";
	copy "$tmpdir/merged.dmp", "$TAXONOMY_LOCAL_LOCATION/";
	print STDERR "Local repository updated.\n";
}

sub init {

	my $parse_flag = shift;

	if (!$TAXONOMY_LOCAL_LOCATION) {
		die "Cannot find the local taxonomy repository! Did you forget to set 'local_repository' in the [Taxonomy] section of smash.conf?\n";
	}
	if (!(-e "$TAXONOMY_LOCAL_LOCATION/nodes.dmp" && -e "$TAXONOMY_LOCAL_LOCATION/names.dmp" && -e "$TAXONOMY_LOCAL_LOCATION/merged.dmp")) {
		print STDERR "Missing NCBI Taxonomy dump files in $TAXONOMY_LOCAL_LOCATION\n";
		update_files();
	}
	if (!%NCBITax2NCBIRank) {
		parse_ncbi_taxonomy_dumps($parse_flag);
	}
	if (!%BergeyTax2BergeyName) {
		parse_rdp_taxonomy_dump();
	}

	# ranks of the levels

	############################################################################
	#
	# Here's the hierarchy of ranks:
	# my @ranks    = ("no rank", qw(domain phylum class subclass order suborder family subfamily supergenus genus));
	# We practically care only about phylum, class, order, family and genus. 
	# The subranks subdomain, subphylum, ..., etc are parsed, but these will 
	# be moved up one level if they are the terminal ranks.
	# Also, we dont care about Incertae Cedis, since it is no information we can use.
	#
	############################################################################

	my @ranks    = qw(norank domain superkingdom kingdom phylum class subclass order suborder family subfamily supergenus genus species);
	%Rank2Num = map {$ranks[$_] => $_} 0..$#ranks;
	$Rank2Num{"no rank"} = $Rank2Num{"norank"};
	$Rank2Num{"no_rank"} = $Rank2Num{"norank"};
	%Num2Rank = map {$_ => $ranks[$_]} 0..$#ranks;
}

sub get_ordered_local_taxonomy_for_id {
	init();
	my $tax_id = shift;
	my %Taxonomy;
	my $node = $tax_id;
	my $level = 0;
	SEARCH:while ($node > 1) {
		my $rank = $NCBITax2NCBIRank{$node};
		$rank =~ s/ /_/g;
		$rank = "${level}__$rank";
		$Taxonomy{$rank} = $NCBITax2NCBIName{$node};
		$node = $NCBITax2NCBIParent{$node};
		$level++;
	}
	return %Taxonomy;
}

sub get_local_taxonomy_for_id {
	init();
	my $id = shift;
	my %taxonomy = get_ordered_local_taxonomy_for_id($id);
	%taxonomy = map {my $k=$_; s/^[0-9]+__//; $_ => $taxonomy{$k}} keys %taxonomy;
	$taxonomy{ncbi_name} = $NCBITax2NCBIName{$id};
	$taxonomy{tax_id}    = $id;
	return %taxonomy;
}

sub get_full_lineage_bergey {
	init();
	my $tax_id = shift;

	if (!defined($BergeyTax2BergeyParent{$tax_id})) {
		die "Cannot find match for $tax_id. Please update bergey taxonomy information.";
	}

	my $candidate = $tax_id;
	my @lineage = ($candidate);
	SEARCH:while ($candidate > 1) {
		$candidate = $BergeyTax2BergeyParent{$candidate};
		push(@lineage, $candidate);
	}
	return @lineage;
}

sub get_full_lineage_ncbi {
	init();
	my $tax_id = shift;
	my $candidate = $tax_id;
	if (!defined($NCBITax2NCBIParent{$candidate})) {
		$candidate = $Old2New{$candidate};
	}

	# When we are here, either there is $NCBITax2NCBIParent for tax_id, or it is remapped using %Old2New.
	# If $candidate doesnt exist now, then $tax_id was missing in NCBI files.

	if (!defined($candidate)) {
		die "Cannot find match for $tax_id. Please update NCBI taxonomy information.";
	}

	my @lineage = ($candidate);
	SEARCH:while ($candidate > 1) {
		$candidate = $NCBITax2NCBIParent{$candidate};
		push(@lineage, $candidate);
	}
	return @lineage;
}

sub is_bacteria_or_archaea {
	init();
	my $tax_id = shift;
	my @lineage = get_full_lineage_ncbi($tax_id);
	foreach my $parent (@lineage) {
		if ($parent == $BACTERIA_ID || $parent == $ARCHAEA_ID) {
			return 1;
		}
	}
	return 0;
}

=item B<get_full_lineage_string($tax_id)>

returns an XML formulated taxonomy lineage.

=cut

sub get_full_lineage_string {
	my %taxonomy = %{$_[0]};

	# taxonomy ranks will be sorted on the position of the rank in this array
	my @levels = qw(tax_id ncbi_name superkingdom kingdom subkingdom superphylum phylum subphylum superclass class subclass infraclass superorder order suborder infraorder parvorder superfamily family subfamily tribe subtribe genus subgenus species_group species_subgroup species subspecies strain);

	my %level = map {$levels[$_] => $_} 0..$#levels;

	# Get the ranked levels
	my @keys = sort {if (!defined($level{$a})) {print STDERR "Unknown $a"}; $level{$a} <=> $level{$b}} grep {$_ !~ /no_rank/} keys %taxonomy; # DEBUG undefined ranks
	#my @keys = sort {$level{$a} <=> $level{$b}} grep {$_ !~ /no_rank/} keys %taxonomy;
	# COOL! WOW FACTOR!
	# Get the unranked levels
	map {				# insert into @keys the
		splice(@keys, $_, 0, "no_rank_$_");
	} sort {			# sorted no_rank
		$a <=> $b
	} map {				# number/position
		(split('_'))[2]
	} grep {			# in unranked
		$_ =~ /no_rank/
	} keys %taxonomy;		# keys

	# Print them
	my $string = "";
	foreach my $key (@keys) {
		my $value = $taxonomy{$key};
		$string .= "$key=\"$value\" ";
	}
	return "<lineage $string/>";
}

sub get_bergey_taxonomic_rank {
	my ($in_tax_id, $required_rank) = @_;
	my $tax_id = $in_tax_id;
	my $required_level = $Rank2Num{$required_rank};
	do {
		if ($Rank2Num{$BergeyTax2BergeyRank{$tax_id}} <= $required_level) {
			return $tax_id;
		}
	} while (($tax_id = $BergeyTax2BergeyParent{$tax_id}) > -1);
	return $in_tax_id;
}

=item B<get_ncbi_taxonomic_rank($tax_id, $rank)>

C<get_ncbi_taxonomic_rank(435590, "genus")> returns 816. If you want the name, then use
C<$NCBITax2NCBIName{816}> to get B<"Bacteroides"> back.

=cut

sub get_ncbi_taxonomic_rank {
	my ($tax_id, $required_rank) = @_;

	$required_rank = "no rank" if $required_rank eq "norank";

	# ranks of the levels

	my $required_level = $Rank2Num{$required_rank};

	# if you dont understand what they want, just give the input back

	if (!defined($required_level)) {
		warn "Do not understand $required_rank";
		return $tax_id;
	}

	my @lineage = get_full_lineage_ncbi($tax_id);
	@lineage = fix_lineage(@lineage);

	
	if ($required_rank eq "species") {

		# hack for Clostridium botulinum
		# since there are many clades under it, i have a priority list. if you hit this, you assign it
		my $hit;
		foreach my $problem (591968, 36826, 36827, 36828, 36829, 36830, 36831, 447213) {
			if (scalar(grep {$problem == $_} @lineage) > 0) {
#warn "$tax_id\t$problem\n";
				return $problem;
			}
		}
		# end


		# hack for wrong species

		my %Remap = (
				# E. coli species
				556266 => 562, # Shigella sp. D9
				457400 => 562, # Escherichia sp. 1_1_43
				469598 => 562, # Escherichia sp. 3_2_53FAA
				457401 => 562, # Escherichia sp. 4_1_40B

				1765   => 1773, # Mycobacterium bovis

				264    => 263, # http://www.bacterio.cict.fr/f/francisella.html#novicida
				234    => 29459,# http://www.bacterio.cict.fr/b/brucella.html#melitensis
				);
		foreach my $problem (keys %Remap) {
			if (scalar(grep {$problem == $_} @lineage) > 0) {
#warn "$tax_id\t".$Remap{$problem}."\n";
				return $Remap{$problem};
			}
		}

		# end
	}


	my $new_id;

	# try to get the right level
	TAXONOMY:foreach my $candidate (@lineage) {
		my $rank = $NCBITax2NCBIRank{$candidate};
		my $level = $Rank2Num{$rank};
		if (defined($level) && $level == $required_level) {
			$new_id = $candidate;
			last TAXONOMY;
		}
	}

	# failed to get the right level
	# so get the closest higher than required-1
	# e.g.,
	# cellular organisms; Bacteria; Firmicutes; Clostridia; Clostridiales; unclassified Clostridiales; unclassified Clostridiales (miscellaneous); butyrate-producing bacterium SR1/5
	# genus returns "unclassified Clostridiales (miscellaneous)"

	if (!$new_id) {
		my $attempted_id;
		TAXONOMY:foreach my $i (0..($#lineage-1)) {
			my $candidate = $lineage[$i];
			my $rank = $NCBITax2NCBIRank{$candidate};
			my $level = $Rank2Num{$rank};
			if (defined($level)) {
				if ($level == $required_level+1) {
					return $lineage[$i+1];
				}
			}
		}
	}

	# so get the closest lower than required+1
	# e.g.,
	# cellular organisms; Bacteria; Firmicutes; Clostridia; Clostridiales; unclassified Clostridiales; unclassified Clostridiales (miscellaneous); butyrate-producing bacterium SR1/5
	# genus returns "unclassified Clostridiales"

	if (!$new_id) {
		my $attempted_id;
		TAXONOMY:foreach my $candidate (@lineage) {
			my $rank = $NCBITax2NCBIRank{$candidate};
			my $level = $Rank2Num{$rank};
			if (defined($level)) {
				if ($level < $required_level) {
					$new_id = $attempted_id;
					last TAXONOMY;
				} else {
					$attempted_id = $candidate;
				}
			}
		}
	}

	if (!$new_id) {
		warn "$tax_id: does not have $required_level\n";
		$new_id = $tax_id;
	}
#warn "$tax_id\t$new_id\n";
	return $new_id;
}

sub fix_lineage {
	my @lineage = @_;

	# first cleanup
	my @out = ();
	foreach my $id (@lineage) {
		if ($id == 384638) { # Bacteroides pectinophilus is in fact an unclassified Lachnospiraceae
			push(@out, $id);
			push(@out, get_full_lineage_ncbi(186928));
			last;
		}
		if ($id == 232721) { # Acidovorax sp. JS42 is in fact a Diaphorobacter
			push(@out, $id);
			push(@out, get_full_lineage_ncbi(238749));
			last;
		}
		if ($id == 103690) { # Nostoc sp. PCC 7120 is in fact an Anabaena
			push(@out, $id);
			push(@out, get_full_lineage_ncbi(1163));
			last;
		}
		if ($id == 28141) { # Cronobacter sakazakii is in fact an Enterobacter
			push(@out, $id);
			push(@out, get_full_lineage_ncbi(547));
			last;
		}

		# updates after calculations for MetaHIT paper
		# will be enabled after paper is cleared
		# if (1 == 0) {
			if ($id == 29348 || $id == 556270) { # Mollicutes bacterium D7 and Clostridium spiroforme DSM 1552 are in fact Coprobacillus
				push(@out, $id);
				push(@out, get_full_lineage_ncbi(100883));
				last;
			}
			if ($id == 518637 || $id == 31971) { # Eubacterium biforme DSM 3989 and Eubacterium dolichum DSM 3991 are in fact an Eubacterium
				push(@out, $id);
				push(@out, get_full_lineage_ncbi(1730));
				last;
			}

			# butyrate-producing bacterium SR1/5 (245020) is actually Blautia(572511)
			# but only 880 bp of 16S
			if ($id == 245020) {
				push(@out, $id);
				push(@out, get_full_lineage_ncbi(572511));
				last;
			}
		# }

		push(@out, $id);
	}

	# then merge
	my @out2 = ();
	foreach my $id (@out) {
		# Merge Escherichia, Shigella, Salmonella, Klebsiella, Citrobacter and Enterobacter
		if (1 == 0) {
			if (scalar(grep {$id == $_} (620,590,570,544,547)) > 0) { # Klebsiella 570 disabled for figure
				push(@out2, get_full_lineage_ncbi(561));
				last;
			}
		}
		push(@out2, $id);
	}
	return @out2;
}

sub get_phylum {
	init();
	my $tax_id = shift;
	my $candidate = $tax_id;
	SEARCH:while (1) {
		my $rank = $NCBITax2NCBIRank{$candidate};
		if ($rank =~ /phylum$/) {
			return $candidate;
		} elsif ($candidate == 1) {
			return undef;
		}
		$candidate = $NCBITax2NCBIParent{$candidate};
	}
}

sub get_common_ancestor_for_two {
	my ($a, $b) = @_;
	my @la = get_full_lineage_ncbi($a);
	my @lb = get_full_lineage_ncbi($b);
	foreach my $anode (@la) {
		foreach my $bnode (@lb) {
			if ($anode == $bnode) {
				return $anode;
			}
		}
	}
	return -1;
}

sub get_common_ancestor_generic {
	my @x = @_;
	my $rca = shift @x; #recent common ancestor
	while (my $node = shift(@x)) {
		$rca = get_common_ancestor_for_two($rca, $node);
	}
	return $rca;
}

sub get_common_ancestor {
	my @x = @_;
	my $count = scalar(@x);
	if ($count == 0) {
		return -1;
	} elsif ($count == 1) {
		return $x[0];
	} elsif ($count == 2) {
		return get_common_ancestor_for_two(@x);
	} else {
		return get_common_ancestor_generic(@x);
	}
}

1;
 
package Smash::Utils::Taxonomy::Deprecated;

use strict;
use warnings;

sub open_genome_tag {
	my $fh     = shift;
	my %genome = %{$_[0]};
	printf $fh "<genome\n\tsource=\"%s\"\n\tname=\"%s\"\n\ttax_id=\"%s\">\n", $genome{source}, $genome{name}, $genome{tax_id};
	print_taxonomy($fh, $genome{taxonomy});
}

sub close_genome_tag {
	my $fh     = shift;
	print $fh "</genome>\n";
}

sub print_taxonomy {
	my $fh       = shift;
	my %taxonomy = %{$_[0]};

	# taxonomy ranks will be sorted on the position of the rank in this array
	my @levels = qw(tax_id ncbi_name superkingdom kingdom subkingdom superphylum phylum subphylum class subclass superorder order suborder superfamily family subfamily tribe subtribe genus subgenus species_group species_subgroup species subspecies strain);

	my %level = map {$levels[$_] => $_} 0..$#levels;
	print $fh "\t<lineage\n";

	# Get the ranked levels
	#my @keys = sort {if (!defined($level{$a})) {print STDERR "Unknown $a"}; $level{$a} <=> $level{$b}} grep {$_ !~ /no_rank/} keys %taxonomy; # DEBUG undefined ranks
	my @keys = sort {$level{$a} <=> $level{$b}} grep {$_ !~ /no_rank/} keys %taxonomy;
	# COOL! WOW FACTOR!
	# Get the unranked levels
	map {				# insert into @keys the
		splice(@keys, $_, 0, "no_rank_$_");
	} sort {			# sorted no_rank
		$a <=> $b
	} map {				# number/position
		(split('_'))[2]
	} grep {			# in unranked
		$_ =~ /no_rank/
	} keys %taxonomy;		# keys

	# Print them
	foreach my $key (@keys) {
		my $value = $taxonomy{$key};
		print $fh "\t\t$key=\"$value\"\n";
	}
	print $fh "\t\t/>\n";
}

our %genome;
our %SeqName2Tax;
our %SeqName2Size;

{
	use XML::Parser;

	############################################
	# XML related
	############################################

	my %buffer;
	my @seqnames;
	my @tax_ids;
	my $parser;      # Parser!!!
	my $tag;
	my $value;                              # Buffers for every element inside <trace>
	my $in_data     = 0;                    # Flag to control appending data inside elements
	my $curr_data   = "";                   # Buffer for data inside elements
	my $first_line  = "<?xml version=\"1.0\"?>";
						# XML Declaration line
	sub parse_genome_xml {
		my $xml_file = shift;

		############################################
		# Set the handlers for the XML refseq
		############################################

		$parser      = new XML::Parser;      # Parser!!!
		$parser->setHandlers(
					XMLDecl => \&xmlDeclaration,
					Start   => \&startElement, 
					End     => \&endElement, 
					Default => \&default
				);

		$parser->parsefile($xml_file);
	}

	sub startElement {
		my( $parseinst, $element, %attrs ) = @_;
		if ($element eq "genome") {
			%buffer = ();
			@seqnames = ();
			foreach my $key (keys %attrs) {
				$buffer{$key} = $attrs{$key};
			}
		} elsif ($element eq "lineage") {
			foreach my $key (keys %attrs) {
				$buffer{$key} = $attrs{$key};
			}
		} elsif ($element eq "sequence") {
			my $def = $attrs{"name"};
			$SeqName2Size{$def} = $attrs{"size"};
			$def =~ s/\.fna//;
			push(@seqnames, $def);
		} else {
			$tag = $element;
		}
	}

	sub endElement {
		my( $parseinst, $element ) = @_;
		if ($element eq "genome") {
			my $tax_id = $buffer{"tax_id"};
			foreach my $key (sort {$a cmp $b} keys %buffer) {
				$genome{$tax_id}{$key} = $buffer{$key};
			}
			foreach my $seq (@seqnames) {
				$SeqName2Tax{$seq} = $tax_id;
			}
			push(@tax_ids, $tax_id);
		} else {
			$genome{$tag} = $value;
		}
		$in_data = 0;
	}

	sub xmlDeclaration {
		my ($parseint, $version, $encoding, $standalone) = @_;
		$first_line = "<?xml";
		if (defined($version)) {
			$first_line .= " version=\"$version\"";
		}
		if (defined($encoding)) {
			$first_line .= " encoding=\"$encoding\"";
		}
		if (defined($standalone)) {
			$first_line .= " standalone=\"$standalone\"";
		}
		$first_line .= "?>";
	}

	sub default {
		my ($parseint, $data) = @_;
		if ($data !~ /\S/) { return;}
		if ($in_data == 1) {
			$curr_data .= $data;
		} else {
			$curr_data  = $data;
		}
		$value = $curr_data;
		$in_data = 1;
		#print "$tag=$data\n";
	}
}

1;

__END__

=back

=head1 DEPENDENCIES

L<Smash::Utils::HTML|Utils::HTML>

=head1 AUTHOR

Mani Arumugam (arumugam@embl.de, http://www.embl.de)

=head1 COPYRIGHT

Copyright (C) 2008-2009. Mani Arumugam and EMBL. All rights reserved.

=cut
