#!/usr/bin/env perl

##################################################################
#   Smash - Simple metagenomic analysis shell                    #
#   Copyright (c) 2008, EMBL-Heidelberg, Germany                 #
#   All Rights Reserved                                          #
#   Send all comments to Mani - arumugam@embl.de                 #
##################################################################

package Smash::Utils::iTOL;

use strict;
use warnings;
use POSIX qw(ceil);
use File::Basename;

use Smash::Global qw(:all);
use Smash::Utils::Taxonomy qw(:all);

my $PROGRESS = \*STDERR;
select($PROGRESS); $| = 1; select(STDOUT);

=begin BOXPLOT

sub make_boxplot_files {
	use Smash::Core;
	use Math::Round;
	my %Median;
	my %Mean;
	my %StDev;
	my %Max;
	my %Min;
	my %First;
	my %Third;
	my %Count;
	my %IQD;
	my %Source; # tax2src mapping
	my %Display;# tax2display mapping
	my %DisplayIds; # display list
	my %Color = ("NCBI" => "#808284", "BROAD" => "#0000FF", "WUGSC" => "#006E3A", "BCM" => "#00ADEF", "JCVI" => "#8E191B", "SC" => "#F26522", "MIX" => "#8CC63F");

	my $smash = new Smash::Core();
	$smash->init();
	my $dbh = $smash->get_refgenomedb_handle();
	my $sth = $dbh->prepare("SELECT source, display_tax_id FROM taxonomy WHERE taxonomy_id=?");

	foreach my $tax_id (keys %Data) {
		$sth->execute($tax_id);
		my ($source, $display_id) = $sth->fetchrow_array();
		$Source{$tax_id} = $source;
		# Keep track of tax_id and display_id - we need them later
		$Display{$tax_id} = $display_id;
		$DisplayIds{$display_id} = 1;
	}
	$sth->finish();
	$dbh->disconnect();

	open(CENTERS, ">$itol_prefix.$refdb.centers.dataset") || die "Cannot open $itol_prefix.$refdb.centers.dataset: $!";
	open(DATASET, ">$itol_prefix.$refdb.dataset") || die "Cannot open $itol_prefix.$refdb.dataset: $!";
	open(COUNTS, ">$itol_prefix.$refdb.counts.dataset") || die "Cannot open $itol_prefix.$refdb.counts.dataset: $!";

	foreach my $tax_id (keys %DisplayIds) {
		# Get all tax ids that map to this display_id
		my @genomes = grep {$Display{$_} eq $tax_id} keys %Display;
		my ($size, $min, $first, $median, $mean, $third, $max, $stdev) = summarize(map {$Data{$_}} @genomes);
		if (!defined($min)) { next; }

		# See how many centers are there
		my %buf = map {$Source{$_} => 1} @genomes;
		my @centers = keys %buf;
		if (scalar(@centers) == 1) {
			print CENTERS join("\t", $tax_id, $Color{$centers[0]}, $centers[0])."\n";
		} else {
			print CENTERS join("\t", $tax_id, $Color{MIX}, "MIX")."\n";
		}
		$Median{$tax_id} = $median;
		$Mean{$tax_id} = $mean;
		$Min{$tax_id} = $min;
		$Max{$tax_id} = $max;
		$First{$tax_id} = $first;
		$Third{$tax_id} = $third;
		$Count{$tax_id} = $size;
		$IQD{$tax_id} = $third-$first;
	}

	my $count = 1;
	#foreach my $item (sort {$Median{$a} <=> $Median{$b}} keys %Median) {}
	foreach my $tax_id (sort {$a <=> $b} keys %Median) {
		my $global_min = 78;
		my $global_max = 100;
		my $scale = 100;
		my $w = 1;
		my ($min, $max, $first, $third, $median) = map {nearest(1, ($_-$global_min)*$scale)+$w+2} map {$Min{$_}, $Max{$_}, $First{$_}, $Third{$_}, $Median{$_}} $tax_id;
		my $length = ($global_max-$global_min)*$scale+2*$w+3; # 2*$w for whiskers and 1 to make it 1-based
		print DATASET join("\t", $tax_id, $length, "RE|1|@{[$min-$w]}|#ffffff|buffer", "RE|$first|$third|#eeeeee|median=".$Median{$tax_id}.";quartile=".$First{$tax_id}."-".$Third{$tax_id}, "RE|@{[$max+$w]}|$length|#ffffff|buffer", "RE|@{[$median-$w]}|@{[$median+$w]}|#000000|median", "RE|@{[$min-$w]}|@{[$min+$w]}|#000000|min", "RE|@{[$max-$w]}|@{[$max+$w]}|#000000|max")."\n";
		$NodesForTree{$tax_id} = 1;
		print COUNTS join("\t", $tax_id, $Count{$tax_id})."\n";
	}

	close(CENTERS);
	close(DATASET);
	close(COUNTS);

	my $dataset_index=1;
	print $CONFIG <<EOF;
dataset${dataset_index}File = $itol_prefix.$refdb.dataset
dataset${dataset_index}Label = $opt_tag Boxplot
dataset${dataset_index}Separator = tab
dataset${dataset_index}Type = domains
dataset${dataset_index}ProtSizeMax = 2000
dataset${dataset_index}Color = \\#FF0000
EOF

	$dataset_index=2;
	print $CONFIG <<EOF;
dataset${dataset_index}File = $itol_prefix.$refdb.centers.dataset
dataset${dataset_index}Label = $opt_tag Centers
dataset${dataset_index}Separator = tab
dataset${dataset_index}Type = colorstrip
dataset${dataset_index}Color = \\#00FF00
dataset${dataset_index}BranchColoringType = both
EOF

	$dataset_index=3;
	print $CONFIG <<EOF;
dataset${dataset_index}File = $itol_prefix.$refdb.counts.dataset
dataset${dataset_index}Label = $opt_tag Count
dataset${dataset_index}Separator = tab
dataset${dataset_index}Type = simplebar
dataset${dataset_index}Color = \\#ffdddd
dataset${dataset_index}BarSizeMax = 1000
EOF
	$smash->finish();
}

sub summarize {
	use PDL;
	use Math::Round;
	my @data_refs = @_;
	my $piddle = qsort pdl(map {@$_} @data_refs);
	my $nelem = nelem($piddle);
	my ($mean, $stdev, $median, $min, $max) = map {(list $_)[0]} stats($piddle);
	my $first = $piddle->at(nearest(1, ($nelem+1)*1/4)-1);
	my $third = $piddle->at(nearest(1, ($nelem+1)*3/4)-1);
	#print join("\t", $nelem, $min, $first, $median, $mean, $third, $max, $stdev)."\n";
	return ($nelem, $min, $first, $median, $mean, $third, $max, $stdev);
}

=cut

sub new {
	my $class = shift;
	my %params = @_;
	my $this  = bless {%params}, $class;

	my $name  = $this->name;
	$this->{DATASET_COUNT} = 0;
	$this->{CONFIG_FILE} = "$name.cfg";
	$this->{TREE_FILE}   = "$name.tree";
	$this->{NCBI_FILE}   = "$name.taxids";
	$this->{COLOR_FILE}  = "$name.colors";

	# fill in defaults if missing

	my %defaults = (
			LOG_SCALE => 0,
			PRINT_SCALE => 0,
			DISPLAY_THRESHOLD => 0,
			NCBI_TAXONOMY_ID => 0,
			NODES_FOR_TREE => {},
			DATASET_COLORS => {},
			PRECISION => 1,
			GRAPH_WIDTH => 100
			);

	foreach my $property (keys %defaults) {
		if (!defined($this->{$property})) {
			$this->{$property} = $defaults{$property};
		}
	}

	return $this;
}

sub init {
	my $this = shift;
	my $name = $this->name;
	my $desc = $this->description;
	my $tree_file = $this->tree_file;
	my $ncbi_file = $this->ncbi_file;
	my $color_file = $this->color_file;
	my $uploadID = $this->uploadID;


	my $config_file = $this->config_file;
	my $treeName = "$desc";
	$treeName .= sprintf("; Time:%s", scalar localtime);

	my $FH;
	open($FH, ">$config_file") || die "Cannot open $config_file: $!";

	print $FH <<EOF;
treeName    = $treeName
uploadID    = $uploadID
projectName = itol_uploader
colorDefinitionFile = $color_file
EOF

	if ($this->ncbi_taxonomy_id) {
		print $FH <<EOF;
ncbiFile    = $ncbi_file
ncbiFormat  = idsCollapsed
assignTaxonomy = 1
EOF
	} else {
		print $FH <<EOF;
treeFile    = $tree_file
treeFormat  = newick
EOF
	}

	$this->{CONFIG_FH} = $FH;
}

sub finish {
	use File::Copy;
	my $this = shift;
	my $nodes_hash     = $this->nodes_for_tree;

	close($this->config_fh);

	# Let iTOL make a tree of NCBI tax_ids

	if ($this->ncbi_taxonomy_id) {
		my $color_file = $this->color_file;
		my $ncbi_file = $this->ncbi_file;
		open(LIST, ">$ncbi_file") || die "Cannot open $ncbi_file: $!";
		open(COLORS, ">$color_file") || die "Cannot open $color_file: $!";
		foreach my $node (keys %$nodes_hash) {
			print LIST "$node\n";
			my @lineage = get_full_lineage_ncbi($node);
			if (@lineage) {
				my %Color = (
						2157   => "#d9fff2",   # Archaea
						201174 => "#ffc4dd",   # Actinobacteria
						976    => "#ffdec7",   # Bacteroidetes
						1239   => "#d9cfff",   # Firmicutes
						1236   => "#b5e0ff",   # Gammaproteobacteria
						1224   => "#e4ffa8",   # Proteobacteria
						#28216  => "#e4ffa8",   # Betaproteobacteria
						#28211  => "#ff6721",   # Alphaproteobacteria
					# eukaryotes
						2759   => "#ff8373",   # Eukaryota
						147545 => "#d9fff2",   # Eurotiomycetes
						147541 => "#ffdec7",   # Dothideomycetes
						147550 => "#d9cfff",   # Sordariamycetes
						);
				foreach my $rank (reverse @lineage) {
					my $color = $Color{$rank};
					if (defined($color)) {
						print COLORS "$node\trange\t$color\t".$NCBITax2NCBIName{$rank}."\n";
					}
				}
			}
		}
		close(COLORS);
		close(LIST);
	} else {
		my $tree_file = $this->tree_file;
		my @nodes     = keys %$nodes_hash;
		if ($this->{RE_SORT}) {
			@nodes = sort {$a <=> $b} @nodes;
		}
		open(LIST, ">$tree_file") || die "Cannot open $tree_file: $!";
		print LIST "(".join(',', @nodes).");\n";
		close(LIST);
	}
	$this->upload();
	my $download = $this->download("svg");
	open(SVG, "<$download") || die "Cannot open $download: $!";
	my $svg = "";
	while (<SVG>) {$svg .= $_};
	close(SVG);

	# remove data labels since we will add it later

	$svg =~ s/<g [^>]*id=\"Rdataset\d+\">.*?<\/g>//sgi;
	$svg =~ s/<g [^>]*id=\"Tdataset\d+\">.*?<\/g>//sgi;
	$svg =~ s/<g id=\'legend\'.*?<\/g>//si;

	# move the color ranges legend up

	$svg =~ s/<g id=\'ranges_legend\' transform=\"translate\(([-\d\.]+),[-\d\.]+\)\">/<g id=\'ranges_legend\' transform=\"translate\(@{[1.5*$1]},0) scale(1.5)\">/si; # scale(2), translate(2*$1,0)

	# add the sample labels above the plot

	my @labels = @{$this->labels};
	foreach my $i (grep {$this->{GRAPH_TYPE}->[$_] eq "barplot"} (1..$this->dataset_count)) {

		# calculate original graph width before iTOL resized it

		my $original_graph_width = 0;
		foreach my $label (@labels) {
			$original_graph_width += $this->{FINAL_GRAPH_WIDTH}->[$i]->{$label};
		}

		# position the new label on top after getting the resized graph widths

		# since we rotate the text, lets use half the space at most, so they are not too close.
		# that's 10*@labels*$font_size in symmetric fonts
		# @labels*2*$font_size will share $picture_size space

		my $font_size = sprintf("%.4f", $this->picture_size/scalar(@labels)/4);
		if ($font_size > 30) {
			$font_size = 30;
		}

		# find out where the graph starts in the svg

		my ($graph) = $svg =~ /<g id="dataset0*${i}" display="block"><line\s+stroke="black" stroke-width="[\d\.]+" stroke-dasharray="[\d\.]+,[\d\.]+" x1="[\d\.]+" y1="[\d\.]+" x2="[\d\.]+" y2="[\d\.]+" style="stroke-width: 1; opacity: [\d\.]+" \/>(.*?)<\/g>/si;
		my ($graph_start_x, $graph_end_y) = (10000, -1000);
		my @rects = $graph =~ /(<rect x=.*? y=.*? height=.*? width=.*? style=.*? \/>)/gsi;
		foreach my $rect (@rects) {
			my ($x, $y, $height) = $rect =~ /<rect x=['"]([\d\.\-]+)['"] y=['"]([\d\.\-]+)['"] height=['"]([\d\.\-]+)['"] width=.*? style=.*? \/>/si;
			if ($x < $graph_start_x) {
				$graph_start_x = $x;
			}
			if ($y+$height > $graph_end_y) {
				$graph_end_y = $y+$height;
			}
		}

		# make the panel

		my $longest_label = (sort {length($b) <=> length($a)} @labels)[0];
		my $panel_width   = 0.75*length($longest_label)*$font_size;
		my $scale_width   = 10;

		($graph_start_x, $panel_width) = map {sprintf("%.8f", $_)} ($graph_start_x, $panel_width);

		# we make the panel by stacking the text and rectangles one on top of another and then rotate them
			# calculate the proportion of this graph's width in the original graph's width and normalize by iTOL picture size

		my $panel_bg = "";
		my $prev_color = "dummy";
		my $panel_color_start = 0;
		my $cumulative_graph_width = 0;
		my $buffer = "";
		foreach my $label (@labels) {
			my $color       = $this->get_color_for_label($label);
			my $graph_width = sprintf("%.8f", $this->picture_size * $this->{FINAL_GRAPH_WIDTH}->[$i]->{$label} / $original_graph_width);

			if ($color ne $prev_color) {
				$buffer .= "<rect x=\"0\" y=\"$panel_color_start\" width=\"$panel_width\" height=\"@{[$cumulative_graph_width-$panel_color_start]}\" style=\"fill:$prev_color;fill-opacity:0.2;\" \/>";
				$panel_color_start = $cumulative_graph_width;
			}
			$cumulative_graph_width += $graph_width;
			$prev_color = $color;
		}
		$buffer .= "<rect x=\"0\" y=\"$panel_color_start\" width=\"$panel_width\" height=\"@{[$cumulative_graph_width-$panel_color_start]}\" style=\"fill:$prev_color;fill-opacity:0.5;\" \/>";
		$panel_bg  = "<g id=\"scale_legend\" transform=\"translate(@{[$graph_start_x-20]}, -$scale_width)\"><text fill=\"#000000\" style=\"font-size:@{[0.75*$font_size]}px; text-anchor:end;\">Scale</text></g>";
		$panel_bg .= "<g id=\"panel_background_$i\" transform=\"translate($graph_start_x, @{[-5*$scale_width]}) rotate(270 0 0)\">".$buffer."</g>\n";

		# add label

		my $panel_text = "";
		$cumulative_graph_width = 0;
		foreach my $label (@labels) {
			my $color       = $this->get_color_for_label($label);
			my $graph_width = sprintf("%.8f", $this->picture_size * $this->{FINAL_GRAPH_WIDTH}->[$i]->{$label} / $original_graph_width);
			$panel_text .= "<text fill=\"$color\" x=\"-2\" y=\"@{[-$cumulative_graph_width-2]}\" style=\"font-size:${font_size}px; font-weight:bold; text-anchor:end;\">$label<\/text>";

			$cumulative_graph_width += $graph_width;
		}
		$panel_text = "<g id=\"panel_text_$i\" transform=\"translate($graph_start_x, @{[-5*$scale_width]}) rotate(90 0 0)\">".$panel_text."</g>\n";

		# add scale

		my $panel_scale = "";
		$cumulative_graph_width = 0;
		foreach my $label (@labels) {
			my $color       = $this->get_color_for_label($label);
			my $graph_width = sprintf("%.8f", $this->picture_size * $this->{FINAL_GRAPH_WIDTH}->[$i]->{$label} / $original_graph_width);
			my $scale       = 0.80*$this->{FINAL_GRAPH_WIDTH}->[$i]->{$label};
			SCALE:foreach my $try (100000, 10, 5, 2, 1, 0.5, 0.2, 0.1, 0.05, 0.02, 0.01, 0.005, 0.002, 0.001) {
				if ($try*int($scale/$try) > 0) {
					$scale = $try*int($scale/$try);
					last SCALE;
				}
			}
			my $scale_px = sprintf("%.8f", $this->picture_size * $scale / $original_graph_width);

			# rescale for display

			$scale /= $this->{SCALE}->[$i];

			# make a scale bar with $scale_width width and value centered on top of it.

			$panel_scale .= "<rect x=\"@{[-2*$scale_width]}\" y=\"@{[-$cumulative_graph_width-$scale_px]}\" width=\"$scale_width\" height=\"$scale_px\" style=\"fill:$color;fill-opacity:0.9;\" \/>";
			my $percent_scale = 1;
			if ($percent_scale == 1) {
				$panel_scale .= "<g transform=\"translate(@{[-2*$scale_width-0.2*$scale_width]}, -$cumulative_graph_width) rotate(270 0 0)\"><text fill=\"$color\" x=\"0\" y=\"0\" style=\"font-size:@{[0.75*$font_size]}px;\">@{[100*$scale]}%<\/text></g>";
			} else {
				$panel_scale .= "<g transform=\"\"><text fill=\"$color\" x=\"@{[-2*$scale_width]}\" y=\"@{[-$cumulative_graph_width]}\" style=\"font-size:@{[0.75*$font_size]}px; font-weight:bold;\">$scale<\/text></g>";
			}

			$cumulative_graph_width += $graph_width;
		}
		$panel_scale = "<g id=\"panel_scale_$i\" transform=\"translate($graph_start_x, 0) rotate(90 0 0)\">".$panel_scale."</g>\n";

		$svg =~ s/(<g id="dataset0*${i}" display="block">)/${panel_bg}${panel_text}${panel_scale}$1/si;
	}

	copy($download, "$download.original");
	open(SVG, ">$download") || die "Cannot open $download: $!";
	print SVG $svg;
	close(SVG);
}

#	make_itol_figure("$itol_prefix.main", $itol_main_config_file, "eps");
#	make_itol_figure("$itol_prefix.piechart", $itol_pie_config_file, "eps");

sub make_itol_figure {
	my ($prefix, $config_file, $format) = @_;
	my $tree_id = upload_itol_tree($config_file);
	download_itol_tree($tree_id, $prefix, "eps", 3);
}

sub upload {
	my $this        = shift;
	my $config_file = $this->config_file;
	my $tree_id;
	my $success = 0;
	my $command = "$SMASH_PERL $SMASH_SCRIPT_LOCATION/iTOL_uploader.pl --config $config_file";
	print "Running: $command\n";
	open(RESPONSE, "$command |") || die "Cannot open pipe to iTOL_uploader: $!";
	while (<RESPONSE>) {
		chomp();
		if (m/^Upload failed/) {
			print "ITOL: $_\n";
			while (<RESPONSE>) {
				print "ITOL: $_";
			}
			return undef;
		} elsif (m/^Upload successful\./) {
			$success = 1;
		} elsif (m/^(\d+)$/ && $success) {
			$tree_id = $1;
		}
		print "ITOL: $_\n";
	}
	$this->{TREE_ID} = $tree_id;
	print "Upload successful. You can view this tree on iTOL using this URL:\n";
	print "http://itol.embl.de/external.cgi?tree=".$this->tree_id."\n";
	return $tree_id;
}

sub download {
	my $this        = shift;
	my $format      = shift;
	my $name        = $this->name;
	my $tree_id     = $this->tree_id;
	my $output_file = "$name.downloaded.$format";
	my $datasets    = join(",", map {"dataset$_"} 1..$this->dataset_count);
	my $command = "$SMASH_PERL $SMASH_SCRIPT_LOCATION/iTOL_downloader.pl --tree $tree_id --format $format --outputFile $output_file --fontSize 20 --displayMode normal --colorBranches 1 --alignLabels 1 --scaleFactor 1 --showBS 1 --BSdisplayValue M0 --BSdisplayType text";
	$command .= " --datasetList $datasets" if $this->dataset_count;
	print "Running: $command\n";
	system($command);
	return $output_file;
}

sub name             {shift->{NAME}}
sub uploadID         {shift->{UPLOAD_ID}}
sub tree_id          {shift->{TREE_ID}}
sub description      {shift->{DESCRIPTION}}
sub labels           {shift->{LABELS}}
sub dataset_name     {shift->{DATASET_NAME}}
sub dataset_count    {shift->{DATASET_COUNT}}
sub dataset_colors   {shift->{DATASET_COLORS}}
sub log_scale        {shift->{LOG_SCALE}}
sub precision        {shift->{PRECISION}}
sub picture_size     {shift->{PICTURE_SIZE}}
sub graph_width      {shift->{GRAPH_WIDTH}}
sub display_threshold{shift->{DISPLAY_THRESHOLD}}
sub nodes_for_tree   {shift->{NODES_FOR_TREE}}
sub ncbi_taxonomy_id {shift->{NCBI_TAXONOMY_ID}}
sub print_scale      {shift->{PRINT_SCALE}}
sub tree_file        {shift->{TREE_FILE}}
sub ncbi_file        {shift->{NCBI_FILE}}
sub merge_level      {shift->{MERGE_LEVEL}}
sub color_file       {shift->{COLOR_FILE}}
sub config_file      {shift->{CONFIG_FILE}}
sub config_fh        {shift->{CONFIG_FH}}

sub merge_taxonomy_ids {
	my $this    = shift;
	my $dataset = shift;
	my $labels  = $this->labels;
	my $merge_level = $this->merge_level;

	my %MergedDataset;

	Smash::Utils::Taxonomy::init();
	my @ids = grep {$_ >= 0} keys %$dataset;
	foreach my $old_id (@ids) {
		my $new_id = get_taxonomic_rank($old_id, $merge_level, 0);
		foreach my $label (@$labels) {
			$MergedDataset{$new_id}{$label} += ($dataset->{$old_id}{$label}||0);
		}
	}
	return \%MergedDataset;
}

sub get_color_for_label {
	my $this   = shift;
	my $label  = shift;
	my $colors = $this->dataset_colors;
	my @available_colors = keys %$colors;
	foreach my $available_color (@available_colors) {
		if ($available_color eq substr($label, 0, length($available_color))) {
			return $colors->{$available_color};
		}
	}
	return "#FF1111";
}

=over 4

=item B<add_dataset(dataset_name, $dataset_hash_reference)

Adds a dataset to the existing iTOL upload job. Passing the reference to the dataset hash is non-destructive.
A local copy is made in the function, so the original is in tact.

=back

=cut

sub add_dataset {
	# All keys %Label are sorted in reverse so that $opt_tag.dataset and $opt_tag.piechart match in their order

	my $this           = shift;
	my %params         = @_;
	my $dataset_name   = $params{NAME};
	my $in_dataset_ref = $params{DATA};
	my $scale          = $params{SCALE} || 1;
	my $graph_type     = $params{GRAPH_TYPE};
	my $keep_top_n     = $params{KEEP_TOP_N};

	my %dataset_hash   = %$in_dataset_ref;
	my $dataset        = \%dataset_hash;

	my $new_dataset    = {};

	my $name           = $this->name;
	my @labels         = @{$this->labels};
	my $dataset_colors = $this->dataset_colors;
	my $log_scale      = $this->log_scale;
	my $precision      = $this->precision;
	my $picture_size   = $this->picture_size;
	my $graph_width    = $this->graph_width;
	my $spacer         = $graph_width / 10;
	my $display_threshold = $this->display_threshold;
	my $nodes_hash     = $this->nodes_for_tree;
	my $dataset_index  = $this->dataset_count+1;
	my $dataset_file   = "$name.$dataset_name.dataset";

	#my %Colors = ("AM-A" => "#808284", "DA-A" => "#0000FF", "ES-A" => "#000000", "FR-A" => "#006E3A", "IT-A" => "#00ADEF", "JP-A" => "#8E191B", "JP-I" => "#F26522", "AM-F" => "#8CC63F");

	# print iTOL config files

	print {$this->config_fh} <<EOF;
dataset${dataset_index}File = $dataset_file
dataset${dataset_index}Label = $dataset_name
dataset${dataset_index}Separator = tab
dataset${dataset_index}Color = \\#FF0000
EOF


	########################
	# Track the max values for each graph to set the width of each graph
	########################

	my %GRAPH_WIDTH = map {$_ => $graph_width} @labels;

	delete $dataset->{-1};
	delete $nodes_hash->{-1};

	# merge tax_ids if necessary

	if ($this->ncbi_taxonomy_id && defined($this->merge_level)) {
		$dataset = $this->merge_taxonomy_ids($dataset);
	}

	# Keep only the top N per sample
	# WARNING: It keeps only topN in the graph... it should keep the topN, but also others that are topN in other samples
	# FIXED: 27.10.2009

	if ($keep_top_n) {
		my %Keep;
		foreach my $label (@labels) {
			my @nodes = sort {$dataset->{$b}{$label} <=> $dataset->{$a}{$label}} grep {defined($dataset->{$_}{$label})} keys %$dataset;
			if (scalar(@nodes) > $keep_top_n) {
				map {$Keep{$_} = 1} @nodes[0..($keep_top_n-1)];
			} else {
				map {$Keep{$_} = 1} @nodes;
			}
		}
		foreach my $node (keys %$dataset) {
			if (!defined($Keep{$node})) {
				delete $dataset->{$node};
			}
		}
	}

	# Throw away very low numbers
	# Find max value of a node in all datasets and delete if it is lower than threshold

	foreach my $node (keys %$dataset) {
		my $max_value_in_sample = 0;
		foreach my $label (@labels) {
			my $value = ($dataset->{$node}->{$label} || 0)*$scale;
			if ($value >= 0) {
				if ($value > $GRAPH_WIDTH{$label}) {$GRAPH_WIDTH{$label} = $value;};
				if ($value > $max_value_in_sample) {$max_value_in_sample = $value;};
			}
		}
		if (!$keep_top_n) {
			if ($max_value_in_sample < $display_threshold*$scale) { delete $dataset->{$node};}
		}
	}

	if ($this->{RE_SORT}) { # re-sort the abundances and forget the actual features. Just keep the ordinal positions
		my @tnodes = keys %$dataset;
		foreach my $label (@labels) {
			my @tmp_nodes = sort {$dataset->{$b}->{$label} <=> $dataset->{$a}->{$label}} grep {defined($dataset->{$_}->{$label})} @tnodes;
			map {$new_dataset->{1+$_}->{$label} = $dataset->{$tmp_nodes[$_]}->{$label}} (0..$#tmp_nodes);
		}
		$dataset = $new_dataset;
	}

	# reassess nodes to display on tree after tree-pruning and editing

	my @nodes = sort keys %$dataset;
	map {$nodes_hash->{$_} = 1} @nodes;

	open(DATASET, ">$dataset_file") || die "Cannot open $dataset_file: $!";

	########################
	# print the header
	########################


	if ($graph_type eq "barplot") {

		####
		# print the labels
		####

		print DATASET "LABELS";
		foreach my $label (@labels) {
			print DATASET "\t$label\t${label}_blank";
		}
		print DATASET "\n";

		####
		# print the colors
		# if there's a color, that will be used. default will be red.
		# colors will be matched by looking for a prefix of the graph label that matches with one of the available color labels
		# e.g., AM-AD-1 will match A, AM, AM-, AM-A, AM-AD, AM-AD-, AM-AD-1. If more than one match is present any one can be used - no guarantees!
		####

		print DATASET "COLORS";
		foreach my $label (@labels) {
			printf DATASET "\t%s\t#ffffff", $this->get_color_for_label($label);
		}
		print DATASET "\n";

		########################
		# print the dataset
		########################

		# round up the maxvalue for display using precision and add the spacer as well

		foreach my $label (@labels) {
			$GRAPH_WIDTH{$label} = $precision*ceil($GRAPH_WIDTH{$label}/$precision) + $spacer;
		}

		if ($log_scale == 1) {
			my %RemappedDataset;
			my $pseudocount = 1;
			foreach my $label (@labels) {
				my $sum = 0;
				foreach my $node (@nodes) {
					my $val = ($dataset->{$node}{$label}||0)*$scale + $pseudocount;
					$RemappedDataset{$node}{$label} = $val;
					$sum += $val;
				}
				foreach my $node (keys %$dataset) {
					$RemappedDataset{$node}{$label} /= $sum;
				}
			}
			my $max = 0;
			foreach my $node (sort @nodes) {
				print DATASET $node;
				foreach my $label (@labels) {
					my $count = $RemappedDataset{$node}{$label};
					my $value = 12+log($count);
					printf DATASET "\t%.4f\t%.4f", $value, 20-$value;
				}
				print DATASET "\n";
			}
		} else {
			# print the data. for each value to be a bar chart, it needs to be padded upto the graphwidth.
			foreach my $node (sort @nodes) {
				print DATASET $node;
				foreach my $label (@labels) {
					my $value = ($dataset->{$node}{$label} || 0)*$scale;
					printf DATASET "\t%.4f\t%.4f", ($value), $GRAPH_WIDTH{$label}-$value;
				}
				print DATASET "\n";
			}
			if ($this->print_scale) {
				my $node = 9606;
				$nodes_hash->{$node} = 1;

				my $value = 10; #scale
				my @labels1 = @labels;
				my $label = shift(@labels1);

				print DATASET $node;
				printf DATASET "\t%.4f\t%.4f", ($value), $GRAPH_WIDTH{$label}-$value;
				foreach my $label (@labels1) {
					my $value = 0;
					printf DATASET "\t%.4f\t%.4f", ($value), $GRAPH_WIDTH{$label}-$value;
				}
				print DATASET "\n";
			}
		}
		close(DATASET);
		print {$this->config_fh} <<EOF;
dataset${dataset_index}Type = multibar
dataset${dataset_index}BarSizeMax = $picture_size
EOF
	} elsif ($graph_type eq "heatmap") {
		####
		# print the labels
		####

		print DATASET "LABELS";
		foreach my $label (@labels) {
			print DATASET "\t$label";
		}
		print DATASET "\n";

		my ($min, $mid, $max);
		my ($min_color, $mid_color, $max_color);

		# write the dataset, and keep track of min and max

		$min = 10000000;
		$max = 0;
		my @values = ();
		foreach my $node (sort @nodes) {
			my $total = 0;
			print DATASET $node;
			foreach my $label (@labels) {
				my $value = ($dataset->{$node}{$label} || 0)*$scale;
				printf DATASET "\t%.4f", $value;
				push(@values, $value) if $value;
			}
			print DATASET "\n";
		}
		@values = sort {$a <=> $b} @values;
		($min, $max) = (0, $values[$#values]);
		$mid = get_median(\@values);
		my $l = 10**int(log($max)/log(10));
		$max  = $l*ceil($max/$l);
		($min_color, $mid_color, $max_color) = ("#0000FF", "#FFFFFF", "#FF0000");

		# if the colors are passed, use them

		if ($params{HEATMAP_COLORS}) {
			my $colors = $params{HEATMAP_COLORS};
			($min, $mid, $max) = sort {$a <=> $b} keys %$colors;
			($min_color, $mid_color, $max_color) = map {$colors->{$_}} ($min, $mid, $max);
		}

		print {$this->config_fh} <<EOF;
dataset${dataset_index}Type = heatmap
dataset${dataset_index}HeatmapBoxWidth = $graph_width
dataset${dataset_index}MinPointValue   = $min
dataset${dataset_index}MidPointValue   = $mid
dataset${dataset_index}MaxPointValue   = $max
dataset${dataset_index}MinPointColor   = \\$min_color
dataset${dataset_index}MidPointColor   = \\$mid_color
dataset${dataset_index}MaxPointColor   = \\$max_color
EOF
	} elsif ($graph_type eq "piechart") {
		####
		# print the labels
		####

		print DATASET "LABELS";
		foreach my $label (@labels) {
			print DATASET "\t$label";
		}
		print DATASET "\n";

		####
		# print the colors
		# if there's a color, that will be used. default will be red.
		####

		print DATASET "COLORS";
		foreach my $label (@labels) {
			printf DATASET "\t%s", $dataset_colors->{$label} || $dataset_colors->{substr($label, 0, 4)} || "#FF1111";
		}
		print DATASET "\n";
		foreach my $node (sort @nodes) {
			my $total = 0;
			foreach my $label (@labels) {
				my $value = ($dataset->{$node}{$label} || 0)*$scale;
				$total += $value;
			}
			printf DATASET "%s\tR%d", $node, sqrt($total);
			foreach my $label (@labels) {
				my $value = ($dataset->{$node}{$label} || 0)*$scale;
				printf DATASET "\t%d", $value;
			}
			print DATASET "\n";
		}
		print {$this->config_fh} <<EOF;
dataset${dataset_index}Type = multibar
dataset${dataset_index}PieRadiusMin = 7
dataset${dataset_index}PieRadiusMax = 13
EOF
	}

	$this->{FINAL_GRAPH_WIDTH}->[$dataset_index] = \%GRAPH_WIDTH;
	$this->{GRAPH_TYPE}->[$dataset_index] = $graph_type;
	$this->{DATASET_COUNT}++;
	$this->{SCALE}->[$dataset_index] = $scale;
}

sub get_median {
	my $list = shift;
	my $n = scalar(@$list);
	my $mid = int($n/2);
	my $median;
	if ($n%2 == 0) {
		$median = ($list->[$mid-1] + $list->[$mid])/2;
	} else {
		$median = $list->[$mid];
	}
	return $median;
}

1;
