#!/usr/bin/env perl

use strict;
use warnings;

use Pod::Usage;
use Bio::SeqIO;
use Smash::Core;
use Smash::Global qw($SMASH_SCRIPT_NAME $SMASH_SCRIPT_LOCATION);
use Smash::CommandLineParser qw(parse_options check_required_options print_options);
use Smash::Analyses::GenePredictor;

##############
# Set up command line parsing
##############

my @allowed  = qw(input=s source=s prefix=s format=s extension=s directory db! help);
my @required = qw(input   source   prefix);

##############
# Parse command line options
##############

my $status;
my $missing;
my %options;

($status, %options) = parse_options(\@allowed);
if ($options{help}) {
	pod2usage(-exitstatus => 0, -verbose => 2);
}
if ($status != 1) {
	pod2usage(-message => "", -exitstatus => 2, -verbose => 1);
}
($status, $missing) = check_required_options(\@required, %options);
if ($status != 1) {
	pod2usage(-message => "$SMASH_SCRIPT_NAME: Missing argument --$missing\n", -exitstatus => 2, -verbose => 1);
}

# get the parameters from commandline

$options{db}= 1 unless defined($options{db});

my $source  = $options{source};
my $prefix  = $options{prefix};
my $input   = $options{input};
my $is_dir  = $options{directory};
my $ext     = $options{extension} || "gbff";
my $format  = $options{format} || "genbank";
my $db      = $options{db};

# init smash object and prepare database connections

my $smash = new Smash::Core();
   $smash->init();

my $dbh;
my ($tax_sth, $seq_sth, $gene_sth, $update_gene_sth, $rna_sth);

# Keep track of rRNAs processed so far to avoid the stupid 23S rRNA split annotations with same GeneID.
# They should have used join() to put them together instead of two rRNA features separated by
# "intervening sequence"

my %ProcessedGene = ();

if ($db) {
	$dbh             = $smash->get_refgenomedb_handle;
	$tax_sth         = $dbh->prepare("INSERT INTO taxonomy(taxonomy_id, source, organism) VALUES(?,?,?);");
	$seq_sth         = $dbh->prepare("INSERT INTO sequence(sequence_id, taxonomy_id, seq_type, mol_type, length, definition) VALUES(?,?,?,?,?,?);");
	$gene_sth        = $dbh->prepare("INSERT INTO gene(external_id, gene_info, sequence_id, type, length, start, end, strand, start_codon, stop_codon, gc) VALUES(?,?,?,?,?,?,?,?,?,?,?);");
	$update_gene_sth = $dbh->prepare("UPDATE gene SET end=? WHERE gene_id=?;");
	$rna_sth         = $dbh->prepare("INSERT INTO mature_rna(gene_id, start, end) VALUES(?,?,?);");
}

# prepare output files

open(PROTEIN, ">$prefix.proteins.fa");
open(GENE, ">$prefix.genes.fa");
open(SEQUENCE, ">$prefix.sequences.fa");

select(STDOUT); $| = 1;
select(STDERR); $| = 1;
select(STDOUT);

# generate list of genbank files, if necessary

my @genbanks;

if ($is_dir) {
	@genbanks = <$input/*.$ext>;
} else {
	@genbanks = ($input);
}

# parse genbank file

foreach my $genbank (@genbanks) {
	my $seqio = Bio::SeqIO->new(-file => $genbank, -format=>$format);
	SEQUENCE: while(my $seq = $seqio->next_seq) {
		my $sequence_id = $seq->id;
		my $definition  = $seq->desc();
		my $length      = $seq->length;
		my $sequence    = $seq->seq();
		my $circular    = $seq->is_circular?"circular":"linear";
		my $tax_id;
		my $organism;
		my $fasta_header = "$sequence_id";
		my $gene_counter = 1;
		if (!$sequence) {
			warn "WARNING: $sequence_id has empty sequence!\n";
			next SEQUENCE;
		}


		FEATURE: foreach my $feature ($seq->get_SeqFeatures) {
			my $primary_tag = $feature->primary_tag;
			my $external_id;   # try GI, GeneID, locus_tag and just a number, in that order!
			my $protein;
			if ($primary_tag eq "source") {
				if ($feature->has_tag("plasmid")) {
					next SEQUENCE;
				}
				if ($feature->has_tag("db_xref")) {
					if ($feature->has_tag("organism")) {
						$organism = ($feature->get_tag_values("organism"))[0];
						if ($organism =~ /phage/i) {
							next FEATURE;
						}
					}
					foreach my $buf ($feature->get_tag_values("db_xref")) {
						($tax_id) = ($buf =~ /taxon:(\d+)/);
					}
				}
				if ($tax_id) {
					$fasta_header = "$tax_id.$fasta_header";
					print join("\t", $genbank, $fasta_header, $length)."\n";

					# eval so that if the tax_id already exists, this will throw an error, but not die!
					if ($db) {
						eval {
							no warnings 'all';
							local $SIG{'__DIE__'};
							$tax_sth->execute($tax_id, $source, $organism);
						};
						eval {
							no warnings 'all';
							local $SIG{'__DIE__'};
							$seq_sth->execute($fasta_header, $tax_id, "chromosome", $circular, $length, $definition);
						};
						next SEQUENCE if $@;
					}
				}
			} elsif (scalar(grep {$primary_tag eq $_} qw(misc_RNA rRNA tRNA CDS)) > 0) {

				# 16S hack again
				if (1 == 0) {
					my $skip = 1;
					if ($feature->has_tag("product")) {
						foreach my $value ($feature->get_tag_values("product")) {
							if ($value =~ /^16S rRNA$/i || $value =~ /^16S ribosomal RNA$/i || $value =~ /^Small Subunit Ribosomal RNA$/i) {
								$skip = 0;
							}
						}
					}
					next FEATURE if $skip;
				}
				# end hack

				###############################################################
				# Some checks that will lead to skipping this feature/sequence
				###############################################################

				# Is this a pseudo gene?

				if ($feature->has_tag("pseudo")) {
					next FEATURE;
				}

				# Sanity check - was the GenBank file incomplete?

				my $location = $feature->location;
				if ($location->end > $length) {
					warn "WARNING: Incomplete GenBank file/record - $genbank:$sequence_id\n";
					next SEQUENCE;
				}

				# Get the gene name (I call it "external_id").
				# Priorities are: 
				#   1. GI
				#   2. GeneID
				#   3. locus_tag
				#   4. gene<number>
				# This goes into the database as tax_id.external_id anyway!

				if ($feature->has_tag("db_xref")) {
					my %dbxref;
					foreach my $buf ($feature->get_tag_values("db_xref")) {
						my ($key, $value) = split(/:/, $buf);
						$dbxref{$key} = $value;
					}
					if ($dbxref{"GI"}) {
						$external_id = $dbxref{"GI"};
					} elsif ($dbxref{"GeneID"}) {
						$external_id = "GeneID:".$dbxref{"GeneID"};
					}
				}
				if (!$external_id) {
					if ($feature->has_tag("locus_tag")) {
						($external_id) = $feature->get_tag_values("locus_tag");
						$external_id = "locus_tag:$external_id";
					}
				}
				$external_id = $external_id || "gene$gene_counter";

				# Get information about the gene from the following fields:
				#    gene, gene_synonym, product, protein_id

				my $gene_info = "";
				my $product;
				if ($feature->has_tag("product")) {
					$product = ($feature->get_tag_values("product"))[0];
				}
				foreach my $tag qw(gene gene_synonym product protein_id) {
					if ($feature->has_tag($tag)) {
						foreach my $value ($feature->get_tag_values($tag)) {
							if ($value ne "hypothetical protein") {
								$gene_info .= "<$tag>$value</$tag>";
							}
						}
					}
				}

				# Hack to fix the 23S split annotations

				if (my $gene_id = $ProcessedGene{"$tax_id.$external_id"}) {

					# Skip the introns

					if ($primary_tag eq "misc_RNA" && $product =~ /intervening sequence/i) {
						warn "NOTE: Skipping $external_id: $primary_tag='$product'\n";
						next FEATURE;
					}

					# Fuse the exons.
					# We know that start < end in both gene and mature_rna tables.
					# So just update the gene table with the higher coordinate, and then add the feature anyway.

					if ($primary_tag eq "rRNA" && $product eq "23S ribosomal RNA") {
						warn "NOTE: Updating $external_id: $primary_tag='$product'\n";
						$update_gene_sth->execute($location->end, $gene_id);
					}
				}

				# Is translation (of course for CDS genes) already in there, or do we have to do it later?

				my $translation_table = 1; # default GenBank trans_table
				if ($feature->has_tag("translation")) {
					$protein = ($feature->get_tag_values("translation"))[0];
				} else {
					if ($feature->has_tag("transl_table")) {
						$translation_table = ($feature->get_tag_values("transl_table"))[0];
						if ($translation_table != 11 && $translation_table != 4) {
							warn "Found non-bacterial/archael translation table for $external_id!\n";
						}
					}
				}

				# Now we are ready to process the feature and add it to the database

				process_gene($primary_tag, $tax_id, $sequence_id, \$sequence, $gene_info, $external_id, $location, $translation_table, $protein);
				$gene_counter++;
			}
		}

		# write the nuc sequence, with tax_id if possible

		print SEQUENCE ">$fasta_header\n";
		print SEQUENCE $smash->pretty_fasta($sequence);

		# write the genome xml
	}
}

# close connections

if ($db) {
	$tax_sth->finish();
	$seq_sth->finish();
	$gene_sth->finish();
	$rna_sth->finish();
	$dbh->commit();
}

$smash->finish();

# close files

close(PROTEIN);
close(GENE);
close(SEQUENCE);

exit(0);

sub process_gene {
	my ($type, $tax_id, $sequence_id, $sequence_ref, $gene_info, $external_id, $location, $translation_table, $protein) = @_;

	my $gene_id; # for the gene created later
	my $complete5 = 0;
	my $complete3 = 0;
	my @cds_locations = ();

	# Check if 5' and 3' complete
	if ($location->start_pos_type() eq "EXACT") {
		$complete5 = 1;
	} elsif ($location->start_pos_type() eq "BEFORE") {
		$complete5 = 0;
	} else {
		die "Cannot handle position type ", $location->start_pos_type(), "\n";
	}
	if ($location->end_pos_type() eq "EXACT") {
		$complete3 = 1;
	} elsif ($location->end_pos_type() eq "AFTER") {
		$complete3 = 0;
	} else {
		die "Cannot handle position type ", $location->end_pos_type(), "\n";
	}

	# get the length of the gene
	my $length = 0;
	if ($location->isa('Bio::Location::SplitLocationI')) {
		foreach my $sublocation ($location->sub_Location) {
			$length += ($sublocation->end-$sublocation->start+1);
			push(@cds_locations, $sublocation->start, $sublocation->end);
		}
	} else {
		$length += ($location->end-$location->start+1);
		push(@cds_locations, $location->start, $location->end);
	}

	# If they are not exact and run outside of the range, get the sequence that's a triplet set
	if ($complete5 == 1 && $complete3 != 1) {
		$cds_locations[$#cds_locations] -= ($length%3);
	} elsif ($complete5 != 1 && $complete3 == 1) {
		$cds_locations[0] += ($length%3);
	}

	# convert them into presence/absence of start/stop codons!
	my $strand = get_strand_char($location->strand);
	my ($start_codon, $stop_codon);
	if ($strand eq "+") {
		($start_codon, $stop_codon) = ($complete5, $complete3);
	} else {
		($start_codon, $stop_codon) = ($complete3, $complete5);
	}

	# Get the transcript 

	my $transcript = "";
	for (my $i=0; $i<@cds_locations; $i+=2) {
		my $cds_start = $cds_locations[$i]; 
		my $cds_end   = $cds_locations[$i+1];
		$transcript  .= substr($$sequence_ref, $cds_start-1, $cds_end-$cds_start+1);
	}
	if ($strand eq "-") {
		$transcript =~ y/ACTG/TGAC/;
		$transcript = reverse($transcript);
	}

	# Fix the frame if CDS open-ended on both sides

	if ($type eq "CDS" && $length%3 != 0) {
		warn $location->seq_id." $external_id not a multiple of 3. Fixing it by finding the longest ORF.\n";
		my $frame   = Smash::Analyses::GenePredictor::get_best_frame($translation_table, $transcript);
		$transcript = substr($transcript, $frame, ($length-$frame) - ($length-$frame)%3);
		if ($strand eq "-") {
			$cds_locations[$#cds_locations] -= $frame;
		} else {
			$cds_locations[0] += $frame;
		}
		my $ptx = Smash::Analyses::GenePredictor::translate($start_codon, $translation_table, $transcript);
	}

	#print join(":", $external_id, $gene_info, $sequence_id, $length, $cds_locations[0], $cds_locations[$#cds_locations], get_strand_char($location->strand), $start_codon, $stop_codon)."\n";

	if ($db) {
		my $gc = $smash->get_gc_percent($transcript, 0, 100);

		# insert into gene/ncgene table

		my $table_external_id = "$tax_id.$external_id";
		$gene_id = $ProcessedGene{$table_external_id};
		if (!defined($gene_id)) {
			$gene_sth->execute($table_external_id, $gene_info, $sequence_id, $type, $length, $cds_locations[0], $cds_locations[$#cds_locations], get_strand_char($location->strand), $start_codon, $stop_codon, $gc);
			$gene_id = $smash->last_refgenomedb_insert_id($dbh);
			$ProcessedGene{$table_external_id} = $gene_id;
		} else {
			print STDERR "NOTE: Skipped $table_external_id\t$type\t$gene_info\n";
		}

		# insert into mature_rna table

		for (my $i=0; $i<@cds_locations; $i+=2) {
			my $cds_start = $cds_locations[$i]; 
			my $cds_end   = $cds_locations[$i+1];
			$rna_sth->execute($gene_id, $cds_start, $cds_end);
		}
	}

	# Get protein
	if ($type eq "CDS" && !$protein) {
		$protein = Smash::Analyses::GenePredictor::translate($start_codon, $translation_table, $transcript);
	}

	print GENE ">$tax_id.$external_id\n";
	print GENE $smash->pretty_fasta($transcript);

	if ($type eq "CDS") {
		print PROTEIN ">$tax_id.$external_id $gene_info\n";
		print PROTEIN $smash->pretty_fasta($protein);
	}

}

sub get_strand_char {
	my $n = shift;
	if ($n < 0) {
		return '-';
	} elsif ($n > 0) {
		return '+';
	} else {
		warn "Strand 0 happened!\n";
		return '.';
	}
}

__END__

=head1 Name

addRefGenomeSequences.pl - Wrapper script to parse genbank/embl files containing 
reference genomes and possibly annotations. It loads the information into Smash 
refgenome database.

=head1 Synopsis

	addRefGenomeSequences.pl [options]

=head1 Options

=over 4

=item B<C<--input>> (required)

genbank file to be parsed, or the directory containing multiple genbank files
to be parsed.

=item B<C<--format>>

format of the input file (supported: genbank, embl)

=item B<C<--directory>>

input is a directory and not a GenBank file. The script will then
look for files with an extension .gbff (or given by C<--extension>) 
in that directory.

=item B<C<--extension>>

file name extension to look for under directory if C<--directory>
is given. (default: gbff)

=item B<C<--source>> (required)

source for the genbank file (e.g., NCBI, HMP, etc)

=item B<C<--prefix>> (required)

prefix for the output files. The following files will be created:

=over 4

=item F<E<lt>prefixE<gt>.sequences.fa>

all the DNA sequences in the genbank file

=item F<E<lt>prefixE<gt>.genes.fa>

all the coding and non-coding genes in the genbank file

=item F<E<lt>prefixE<gt>.proteins.fa>

protein sequences of the coding genes in the genbank file (translation
in the genbank file is used if available, otherwise the CDS regions are
translated by this script).

=back

=item B<C<--help>>

Prints this manual.

=back

=head1 Description

=head1 Author

Mani Arumugam (arumugam@embl.de, http://www.embl.de)

=head1 Copyright

Copyright (C) 2008-2009. Mani Arumugam and EMBL. All rights reserved.

=cut
