#! /usr/bin/env perl

use strict;
use warnings;
use Pod::Usage;
use Smash::Core;
use Smash::Global qw($SMASH_SCRIPT_NAME $SMASH_SCRIPT_LOCATION);
use Smash::CommandLineParser qw(parse_options check_required_options);
use File::Temp;

##############
# Set up command line parsing
##############

my @allowed  = qw(genepred=s string=s help); # arguments I expect
my @required = qw(genepred string);            # arguments I require

##############
# Parse command line options
##############

my $status;
my $missing;
my %options;

($status, %options) = parse_options(\@allowed);
if ($options{help}) {
	pod2usage(-exitstatus => 0, -verbose => 2);
}
if ($status != 1) {
	pod2usage(-message => "", -exitstatus => 2, -verbose => 1);
}
#print_options(%options);
($status, $missing) = check_required_options(\@required, %options);
if ($status != 1) {
	pod2usage(-message => "$SMASH_SCRIPT_NAME: Missing argument --$missing\n", -exitstatus => 2, -verbose => 1);
}

##############
# do the mapping
##############

my $string       = lc($options{string});
my $genepred     = $options{genepred};
my $smash        = new Smash::Core(GENEPRED => $genepred);
   $smash->init();
my $genepred_id  = $smash->get_id_by_name("gene_prediction", $genepred);
my $genepred_dir = $smash->genepred_dir($genepred);
my $external_dir = $smash->get_smash_conf_value("data_dir")."/external";

# files used for og mapping

my $blast_file   = "$genepred_dir/$genepred.$string.blastp";
my $orthgrp_file = "$external_dir/${string}_final_orthgroups.txt";
my $plength_file = "$external_dir/${string}_protein_lengths.txt";
my $og_map_file  = "$genepred_dir/$genepred.eggnogmapping.txt";
my $contig2gene  = new File::Temp();

# make best hit file

$smash->execute("$SMASH_SCRIPT_LOCATION/find_best_hit.pl -i $blast_file > $blast_file.besthit");

# make contig gene mapping file

{
	my $dbh = $smash->get_db_handle;
	my $sth = $dbh->prepare("SELECT c.external_id, g.external_id, start, end, strand FROM contig c INNER JOIN gene g USING (contig_id) WHERE gene_prediction_id=$genepred_id");
	my ($contig, $gene, $start, $end, $strand);
	$sth->execute();
	$sth->bind_columns(\$contig, \$gene, \$start, \$end, \$strand);
	while ($sth->fetch()) {
		print $contig2gene join("\t", $contig, $gene, $start, $end, ($strand eq "+")?"forward":"reverse")."\n";
	}
	close($contig2gene);
}

# run og mapping

$smash->execute("$SMASH_SCRIPT_LOCATION/og_mapping.py -b $blast_file -t $blast_file.besthit -g $orthgrp_file -m 0.5 -s 60 -r $contig2gene -p $plength_file > $og_map_file.preliminary");

# clean up

$smash->execute("egrep -v '^#| NONE ' $og_map_file.preliminary > $og_map_file");

# done!

$smash->finish();

exit(0);

__END__

=head1 doOrthologMapping.pl

Script to map proteins to STRING orthologous groups (OGs)
using BLASTP results against STRING.

=head1 Synopsis

	doOrthologMapping.pl [options]

=head1 Options

=over 4

=item B<C<--genepred>>

Name of geneprediction whose proteins should be mapped.

=item B<C<--string>>

Version of the string database to be used.

=item B<C<--help>>

Prints this manual.

=back

=head1 Description

B<doOrthologMapping.pl> is a wrapper script that maps predicted proteins 
in a metagenome to STRING orthologous groups. It is run as:

	doOrthologMapping.pl --genepred=MC20.MG10.AS2.GP1 --string=string8

=head1 Required files

B<doOrthologMapping.pl> requires the following 
files (assuming string version 8; replace 8 with the correct version):

=over 4

=item 1. BLAST results

File containing results from BLASTing the predicted proteins against 
STRING proteins. Please download the STRING proteins from 
L<http://string.embl.de> if you haven't already. Check the "Downloads"
section for a flat file containing all STRING proteins. As of writing this
manual, the latest version of STRING is 8.2. Changes in the minor 
version (after the decimal point) usually do not affect the protein set
(except between 8.0 and 8.1). Only the major version is required by 
this script so that it uses the correct information to map proteins.

This file is expected to be in the gene prediction directory. Once you have
the results of BLAST, please move it to that directory. To see the
location of this directory for a given gene prediction, run:

	perl showLocations.pl --item=MC10.MG23.AS1.GP2

=item 2. STRING orthologous group information

These files can be downloaded from SMASH website. They are available at
L<http://www.bork.embl.de/software/smash/downloads/>. Here are direct links to the
STRING7 orthgroup file (L<http://www.bork.embl.de/software/smash/downloads/og_mapping/string7_final_orthgroups.txt>)
and
STRING8 orthgroup file (L<http://www.bork.embl.de/software/smash/downloads/og_mapping/string8_final_orthgroups.txt>).
Copy these files to the SMASH external resource directory. To see the 
location of this directory, run:

	perl showLocations.pl --external


=item 3. STRING protein lengths

Tab delimited file containing names of proteins in STRING and their 
lengths - one protein per line like so: 

	367928.BAD_1369	328
	299033.Lreu_1369	338

If you have the fasta file containing STRING proteins, this is easy 
to generate. If you don't know how to do it, here's a quick solution
for you using L<FAlite|"lib/FAlite">. If you have SMASH configured 
already, FAlite will be in your PERL library path. Copy and paste the
following lines to a perl script (let us call it F<get_length.pl>).

	#!/usr/bin/env perl
	use strict;
	use warnings;
	use FAlite; 
	my $f = new FAlite(\*STDIN);
	while (my $e=$f->nextEntry) {
		my @header = split(/\s/, $e->def); # Thanks to Youssef Darzi for the fix!
		print substr($header[0], 1)."\t".length($e->seq)."\n";
	} 

Then run the following command:

	cat protein.sequences.v8.2.fa | perl get_length.pl > string8_protein_lengths.txt

If you have the gzipped fasta file, 

	zcat protein.sequences.v8.2.fa.gz | perl get_length.pl > string8_protein_lengths.txt

Once you generate this file, move this to the SMASH external 
resource directory. See above to get the location of this
directory.

=back

=cut

=head1 Author

Mani Arumugam (arumugam@embl.de, http://www.embl.de)

=head1 Copyright

Copyright (C) 2008-2009. Mani Arumugam and EMBL. All rights reserved.

=cut
