#! /usr/bin/env perl

use strict;
use warnings;
use Pod::Usage;
use Cwd;
use File::Path;
use File::Copy;
use Smash::Core;
use Smash::Global qw($SMASH_SCRIPT_NAME);
use Smash::CommandLineParser qw(parse_options check_required_options);

select(STDERR); $| = 1; select(STDOUT);

##############
# Set up command line parsing
##############

my @allowed  = qw(type=s name=s version=s cluster=s host=s license=s current help);
my @required = qw(type name version);

##############
# Parse command line options
##############

my $status;
my $missing;
my %options;

($status, %options) = parse_options(\@allowed);
if ($options{help}) {
	pod2usage(-exitstatus => 0, -verbose => 2);
}
if ($status != 1) {
	pod2usage(-message => "", -exitstatus => 2, -verbose => 1);
}
#print_options(%options);
($status, $missing) = check_required_options(\@required, %options);
if ($status != 1) {
	pod2usage(-message => "$SMASH_SCRIPT_NAME: Missing argument --$missing\n", -exitstatus => 2, -verbose => 1);
}

##############
# Handle command line options
# (except checking for presence of required args, which has already 
# been done by parse_options)
##############

if ($options{cluster} && $options{host}) {
	warn "--cluster and --host are specified: using only --cluster\n";
}

my $type = $options{type};
my $software_name = $options{name};
my $version = $options{version};

if ($software_name eq "GeneMark" && !defined($options{license})) {
	pod2usage(-message => "$SMASH_SCRIPT_NAME: --name=GeneMark requires --license\n", -exitstatus => 2, -verbose => 1);
}

##############
# figure out the source and destination directories
##############

my $source = Cwd::cwd;
my $smash = new Smash::Core();
$smash->init();
my ($machine) = grep {defined($_)} ($options{cluster}, $options{host}, "generic");
my $pkg_dir = sprintf("%s/%s/%s/%s/%s/%s", $smash->data_dir, $smash->get_smash_conf_value("software_dir"), $machine, $type, $software_name, $version);

###############
# Arachne
# extra files
###############

if ($software_name eq "Arachne" && lc($version) eq "extra") {
	my $now = time;
	my $Arachne_data = sprintf("%s/%s/Assembler/Arachne/Arachne_data", $smash->data_dir, $smash->get_smash_conf_value("workspace_dir"));

	map {mkpath("$Arachne_data/$_", {mode => $smash->file_perm})} qw(dtds vector e_coli e_coli_transposons);
	map {open(FP, ">$Arachne_data/$_/contigs.fasta"); close(FP);} qw(dtds vector e_coli e_coli_transposons);
	map {copy $_, "$Arachne_data/$_"}                             qw(dtds/configuration.dtd vector/contigs.fasta);
	copy "L3191/reads_config.xml", $Arachne_data;
	exit(0);
}

##############
# special case - Celera
##############

if ($software_name eq "Celera") {
	my $target = `uname`;
	chomp($target);
	if ($target eq "Linux") {
		$target = `uname -m`;
		chomp($target);
		my $subdir;
		if ($target eq "x86_64") {
			$subdir="Linux-amd64";
		} else {
			$subdir="Linux-i686";
		}
		$source = "$source/$subdir" if $subdir;
		$pkg_dir = "$pkg_dir/$subdir" if $subdir;
	} else {
		die "Smash can only install external software on Linux platforms!";
	}
}

##############
# copy files
##############

print STDERR "cp -R $source/* $pkg_dir/\n";
print STDERR "Copying the contents of $source to $pkg_dir ...";
mkpath($pkg_dir, {mode => 0770});
$smash->execute("cp -R $source/* $pkg_dir/");
print STDERR " done\n";

###############
# special cases
###############

###############
# GeneMark
# copy license
###############

if ($software_name eq "GeneMark") {
	copy $options{license}, "$pkg_dir/gm_key";
}

##############
# make this current, if applicable
##############

if ($options{current}) {
	my $current = sprintf("%s/%s/%s/%s/%s/%s", $smash->data_dir, $smash->get_smash_conf_value("software_dir"), $machine, $type, $software_name, "current");
	unlink $current;
	symlink $version, $current;
}

exit(0);


__END__


=head1 Name

installSoftware.pl - Smash driver script for installing external software

=head1 Synopsis

	installSoftware.pl [options]

=head1 Options

=over 4

=item B<C<--type>> (required)

type of software being installed (Assembler or GenePredictor)

=item B<C<--name>> (required)

name of the software program (Arachne|Celera|Forge for Assembler, and GeneMark|MetaGene for GenePredictor)

=item B<C<--version>> (required)

version being installed

=item B<C<--cluster>>

name of cluster 

=item B<C<--host>>

name of host

=item B<C<--license>>

location of license file, required only for name=GeneMark

=item B<C<--help>>

Prints this manual.

=back

=head1 Description

B<installSoftware.pl> is a wrapper script to install external software into Smash repository. It currently supports
the following programs:

=over 4

=item Assemblers:

Arachne, Celera

=item Gene predictors:

GeneMark, MetaGene

=back

A normal execution of this script would be:

	installSoftware.pl --type=Assembler --name=Arachne \
	    --version=3.1 --host=host123

Please note that "installation" here refers to installing these packages in Smash. Running this script does not
compile and build the software. That must have been done by the user prior to running this script. Please refer
to the manual of the respective software if you need help with compiling and building them.

=head1 Author

Mani Arumugam (arumugam@embl.de, http://www.embl.de)

=head1 Copyright

Copyright (C) 2008-2009. Mani Arumugam and EMBL. All rights reserved.

=cut
