#! /usr/bin/env perl

##################################################################
#   Smash - Simple metagenomic analysis shell                    #
#   Copyright (c) 2008, EMBL-Heidelberg, Germany                 #
#   All Rights Reserved                                          #
#   Send all comments to Mani - arumugam@embl.de                 #
##################################################################

use strict;
use warnings;
use Pod::Usage;
use Getopt::Long;
use File::Basename;
use FAlite;
use Smash::Global qw($SMASH_SCRIPT_NAME);
use Smash::CommandLineParser qw(parse_options check_required_options);
use Smash::Analyses::Assembler;
use Smash::Analyses::GenePredictor;
use Smash::Utils::GFFlite;
use Smash::Databases::MetaGenomeDB::Loader qw(load_smash);

##############
# Set up command line parsing
##############

my @allowed  = qw(metagenome=s assembler=s version=s parameters=s contig_fasta=s contig_gff=s scaffold_gff=s gene_gff=s genepred=s genepred_version=s genepred_parameters=s help); # arguments I expect
my @required = qw(metagenome assembler version contig_fasta contig_gff);                       # arguments I require

##############
# Parse command line options
##############

my $status;
my $missing;
my %options;

($status, %options) = parse_options(\@allowed);
if ($options{help}) {
	pod2usage(-exitstatus => 0, -verbose => 2);
}
if ($status != 1) {
	pod2usage(-message => "", -exitstatus => 2, -verbose => 1);
}
#print_options(%options);
($status, $missing) = check_required_options(\@required, %options);
if ($status != 1) {
	pod2usage(-message => "$SMASH_SCRIPT_NAME: Missing argument --$missing\n", -exitstatus => 2, -verbose => 1);
}

##############
# Handle command line options
# (except checking for presence of required args, which has already 
# been done by parse_options)
##############

my $metagenome   = $options{metagenome};
my $contig_fasta = $options{contig_fasta};
my $contig_gff   = $options{contig_gff};
my $scaffold_gff = $options{scaffold_gff};
my $fake_scaf    = !defined($scaffold_gff);
my $gene_gff     = $options{gene_gff};

############################################
# Instantiate an external assembler
############################################

$options{OPTIONS} = $options{parameters};
my $assembler = new Smash::Analyses::Assembler::external(map {uc($_) => $options{$_}} keys %options);
$assembler->init(); 
$assembler->prepare(); # Get a new name, init the directory structure etc.

############################################
# Get the required files and locations and
# other parameters
############################################

my $assembly     = $assembler->name;
my $assembly_dir = $assembler->assembly_dir;
my $contig2read  = "$assembly_dir/$assembly.contig2read.gff";
my $scaf2contig  = "$assembly_dir/$assembly.scaf2contig.gff";
my $contig_fa    = "$assembly_dir/$assembly.contigs.fa";

my %External2Smash;

############################################
# Process the read fasta file and make
# a contig per read and dump it in CONTIGS
# and CONTIG2READ files
############################################

open(IN_CONTIGS, "<$contig_fasta") || die "Cannot open $contig_fasta: $!";
open(CONTIG2READ, ">$contig2read") || die "Cannot open $contig2read: $!";
open(SCAF2CONTIG, ">$scaf2contig") || die "Cannot open $scaf2contig: $!";
open(CONTIG_FA, ">$contig_fa") || die "Cannot open $contig_fa: $!";

############################################
# Read the contig fasta file and rename the 
# contigs using SMASH standards.
############################################

my $fasta = new FAlite(\*IN_CONTIGS);
my $count = 0;
while (my $entry = $fasta->nextEntry) {
	$count++;
	my $contig_id = "$assembly.C$count";
	my $contig_length = length($entry->seq);
	my $contig_name = $assembler->process_fasta_header($entry->def);
	print CONTIG_FA ">$contig_id\n";
	print CONTIG_FA $assembler->pretty_fasta($entry->seq);
	$External2Smash{$contig_name} = $contig_id;
	if ($fake_scaf) {
		my $scaf_id   = "$assembly.S$count";
		print SCAF2CONTIG join("\t", $scaf_id, "$assembler", "contig", 1, $contig_length, $contig_length, "+", ".", "contig \"$contig_id\";\n");
	}
}
close(IN_CONTIGS);

############################################
# Read the contig2read gff file and remap
# the names then make the proper contig2read
# file in SMASH repository
############################################

open(GFF, "<$contig_gff") || die "Cannot open $contig_gff: $!";
my $gff = new Smash::Utils::GFFlite(\*GFF);
while (my $f = $gff->nextFeature) {
	print CONTIG2READ join("\t", $External2Smash{$f->seqname}, "$assembler", "read", $f->start, $f->end, $f->score, $f->strand, $f->frame, $f->flattened_attributes)."\n";
}
close(GFF);

if ($scaffold_gff) {
	open(GFF, "<$scaffold_gff") || die "Cannot open $scaffold_gff: $!";
	my $gff = new Smash::Utils::GFFlite(\*GFF);
	my $count = 0;
	while (my $f = $gff->nextFeature) {
		$count++;
		my $scaf_id   = "$assembly.S$count";
		my $contig = $f->get_attribute("contig");
		$f->{ATTRIBUTES}->{contig} = $External2Smash{$contig};
		print SCAF2CONTIG join("\t", $scaf_id, "$assembler", "contig", $f->start, $f->end, $f->score, $f->strand, $f->frame, $f->flattened_attributes)."\n";
	}
	close(GFF);
}

close(CONTIG_FA);
close(CONTIG2READ);
close(SCAF2CONTIG);
$assembler->finish();

print "********************************************************\n";
print "   Assembly id assigned for this assembly: $assembly\n";
print "********************************************************\n";

load_smash("assembly", (ASSEMBLY => $assembly));

# Load gene annotations as well!

if (defined($gene_gff)) {
	$options{OPTIONS}  = $options{genepred_parameters};
	$options{VERSION}  = $options{genepred_version};
	$options{ASSEMBLY} = $assembly;
	my $predictor = new Smash::Analyses::GenePredictor::external(map {uc($_) => $options{$_}} keys %options);
	$predictor->init();
	my $prediction   = $predictor->name;
	my $genepred_dir = $predictor->genepred_dir($prediction);
	my $contig2gene  = "$genepred_dir/$prediction.contig2gene.gff";
	open(CONTIG2GENE, ">$contig2gene") || die "Cannot open $contig2gene: $!";
	open(GENE_GFF, "<$gene_gff") || die "Cannot open $gene_gff: $!";
	my $gff = new Smash::Utils::GFFlite(\*GENE_GFF);
	while (my $f = $gff->nextFeature) {
		my $seqname = $External2Smash{$f->seqname};
		$f->set_property("SEQNAME", $seqname);
		$f->set_property("SOURCE", "$predictor");
		$seqname =~ s/^${assembly}//;
		$f->set_attribute("gene_id", "$prediction$seqname.".$f->get_attribute("gene_id"));
		Smash::Utils::GFF::Feature::print_feature_gff($f, \*CONTIG2GENE);
	}
	close(GENE_GFF);
	close(CONTIG2GENE);
	$predictor->finish();

	print "********************************************************\n";
	print "   Gene Prediction id assigned for this assembly: $prediction\n";
	print "********************************************************\n";
	load_smash("geneprediction", ('GENEPRED' => $prediction));
	print "<output>success</output>\n";
}

exit(0);

############################################
# End Execution
############################################

__END__

=head1 Name

loadExternalAssembly.pl - Wrapper script to load an external assembly into SMASH repository

=head1 Synopsis

	loadExternalAssembly.pl [options]

=head1 Options

=over 4

=item B<C<--metagenome>>

(required) name of the metagenome this assembly belongs to. This metagenome must be present in the repository, and reads
should already have been loaded.

=item B<C<--assembler>>

(required) name of the assembly software used to generate this assembly outside of SMASH. 

=item B<C<--version>>

(required) version of the assembly software used to generate this assembly outside of SMASH. 

=item B<C<--parameters>>

(optional) special parameters used to generate this assembly outside of SMASH. 

=item B<C<--contig_fasta>>

(required) fasta file containing the assembled contig sequences

=item B<C<--contig_gff>>

(required) tab-delimited GFF file containing read to contig mapping in the following format:

	<contig_id>  <assembler>  read  <start>  <end>  <contig_length>  <strand>  .  \
	read "<read_name>"; [mate_pair "<mate_name>"; [contig "<contig_name>"; insert_size "<insert_size>";]]

E.g.,

	contig1	Newbler	read	131	259	2023	+	.	\
	read "MC99.MG1.ABC.y"; mate_pair "MC99.MG1.ABC.z"; contig "contig1"; insert_size 1000;
	contig1	Newbler	read	198	363	2023	+	.	\
	read "MC99.MG1.XYZ.y";
	contig1	Newbler	read	910	1030	2023	+	.	\
	read "MC99.MG1.ABC.z"; mate_pair "MC99.MG1.ABC.y"; contig "contig1"; insert_size 1000;

=item B<C<--scaffold_gff>>

GFF file containing contig to scaffold mapping in a format similar to the read to contig mapping

=item B<C<--gene_gff>>

GFF file containing gene coordinates in the standard GFF format

=item B<C<--help>>

Prints this manual.

=back

=head1 Description

B<loadExternalAssembly.pl> is a wrapper script to load an external assembly into SMASH
repository and database. 

A typical use of B<loadExternalAssembly.pl> follows B<loadMetaGenome.pl>, like so:

	loadExternalAssembly.pl --metagenome=MC99.MG1 --contig_fasta=assembly.fasta \
	    --contig_gff=contigs.gff --scaffold_gff=scaffolds.gff

If you dont have scaffolding information, then you could use

	loadExternalAssembly.pl --metagenome=MC99.MG1 --contig_fasta=assembly.fasta \
	    --contig_gff=contigs.gff

The GFF file can have contig names generated by the external assembly program. The script
will rename the contigs to SMASH format. However, the contig names in the contig_fasta 
file and the contig_gff files MUST match, since the script will match them accordingly 
the contig names.

Some assemblies have scaffold information, and some don't. If you do have scaffold 
information, then you must create a GFF file similar to the one above, but to map 
contigs onto scaffold. If you dont have scaffold information, then do not create a 
scaffold_gff file. Without a C<--scaffold_gff> option, a "fake" scaffold is created for each contig. 

B<Optionally>, you can also load an external gene prediction along with the external assembly. We
recommend loading the external assembly into SMASH first, and using the contig FASTA file from the 
repository to make gene prediction, and then load the genes using L<loadGenePrediction.pl|loadGenePrediction>,
but if that is not possible here is a way to load both the assembly and gene prediction together.

=head1 Author

Mani Arumugam (arumugam@embl.de, http://www.embl.de)

=head1 Copyright

Copyright (C) 2008-2009. Mani Arumugam and EMBL. All rights reserved.

=cut
